"""
Base database model
"""
from datetime import datetime, timezone
from pyjolt.database.sql import DeclarativeBaseModel
from sqlalchemy import DateTime
from sqlalchemy.orm import mapped_column, Mapped

class DatabaseModel(DeclarativeBaseModel):
    """Base for all database models"""

    __abstract__ = True
    __db_name__ = "sql_database"

    id: Mapped[int] = mapped_column(primary_key=True)
    created_at: Mapped[datetime] = mapped_column(DateTime(timezone=True),
                                                 default=lambda: datetime.now(timezone.utc),
                                                nullable=False)
    
    class Meta(DeclarativeBaseModel.Meta):
        pass



"""
Base response model
"""
from typing import Optional
from pydantic import BaseModel, Field

class BaseResponse(BaseModel):
    """Base"""
    message: Optional[str] = Field("Request successful")
    status: Optional[str] = Field("success")
    data: Optional[BaseModel] = Field(None, description="Data payload - a pydantic object")

class ErrorResponseSchema(BaseModel):
    """Generic error response schema"""
    message: str
    status: str
    data: Optional[dict] = None

class QueryResponseSchema(BaseModel):
    """Query response schema"""
    items: list[BaseModel]
    total: int
    page: int
    pages: int
    per_page: int
    has_next: bool
    has_prev: bool



"""Project related schemas"""
from datetime import datetime
from typing import Optional, TYPE_CHECKING, Self
from pydantic import BaseModel

from .base_schema import QueryResponseSchema, BaseResponse

if TYPE_CHECKING:
    from app.api.models.project import Project

class ProjectQuerySchema(BaseModel):
    """Schema for project queries"""
    title_slv: Optional[str] = None
    title_eng: Optional[str] = None
    keywords_slv: Optional[str] = None
    keywords_eng: Optional[str] = None
    leader: Optional[str] = None
    start_date: Optional[datetime] = None
    end_date: Optional[datetime] = None
    researchers: Optional[str] = None

    page: int = 1
    per_page: int = 10

class ProjectOutSchema(BaseModel):
    """Project out schema"""

    id: int
    title_slv: Optional[str] = None
    title_eng: Optional[str] = None
    code: Optional[str] = None
    abstract_slv: Optional[str] = None
    abstract_eng: Optional[str] = None
    keywords_slv: Optional[str] = None
    keywords_eng: Optional[str] = None
    leader: Optional[str] = None
    duration: Optional[str] = None
    researchers: Optional[str] = None

    @classmethod
    def from_model(cls, project: "Project") -> Self:
        """Creates schema from project model"""
        return cls(
            id=project.id,
            code=project.code,
            title_slv=project.title_slv,
            title_eng=project.title_eng,
            abstract_slv=project.abstract_slv,
            abstract_eng=project.abstract_eng,
            keywords_slv=project.keywords_slv,
            keywords_eng=project.keywords_eng,
            leader=project.leader,
            duration=project.duration,
            researchers=project.researchers
        )

class ProjectResponseSchema(BaseResponse):
    """Project response schema"""
    data: ProjectOutSchema

class ProjectsQueryDataSchema(QueryResponseSchema):
    """Publication API base query response"""
    items: list[ProjectOutSchema]

class ProjectsQueryResponseSchema(BaseResponse):
    """Publications API query response"""
    data: ProjectsQueryDataSchema



"""Development utilities"""
import re
from typing import Any
import requests
from requests import Response
from pyjolt.database.sql import AsyncSession
from pyjolt.cli import CLIController, command, argument
from faker import Faker

from app.api.models.post import Post
from app.api.models.publication import Publication
from app.extensions import db

from .publications_doi import dois

def slugify(s: str) -> str:
    """
    Slugyfies text
    """
    s = s.lower().strip()
    s = re.sub(r"[^\w\s-]", "", s)      # drop non-word chars
    s = re.sub(r"[\s_-]+", "-", s)      # collapse to single hyphens
    s = s.strip("-")
    return s or "post"

class DevUtilities(CLIController):
    """Simple dev utility controller"""

    @command("create-posts")
    @argument("number", int)
    @db.managed_session_for_cli
    async def create_fake_posts(self, number: int, session: AsyncSession):
        """Creates any number of fake posts"""
        fake: Faker = Faker()
        posts: list[Post] = []
        for _ in range(number):
            title_en = fake.sentence().rstrip(".")
            title_sl = fake.sentence().rstrip(".")
            content_en = "\n\n".join(fake.paragraphs(nb=5))
            content_sl = "\n\n".join(fake.paragraphs(nb=5))
            candidate_slug = slugify(title_en)

            post = Post(
                title_eng=title_en,
                title_slv=title_sl,
                slug=candidate_slug,
                content_eng=content_en,
                content_slv=content_sl,
                active=True,
                tags_list=",".join([t.lower() for t in fake.words(nb=5)]),
                author_id=1
            )
            posts.append(post)
        session.add_all(posts)
        print(f"Created {number} of posts")

    @command("fetch-publications")
    @db.managed_session_for_cli
    async def fetch_publications(self, session: AsyncSession):
        """
        Fetched all publications data from DOI list
        Uses api.crossref.com
        """
        base_url: str = self.app.get_conf("CROSSREF_API_URL")
        pubs: list[Publication] = []
        for doi in dois:
            try:
                response: Response = requests.get(f"{base_url}/{doi}", timeout=20)
                if response.status_code != 200:
                    print(f"Failed to fetch (status={response.status_code}): ", doi)
                    continue
                data: dict[str, Any] = response.json()
                publication = Publication.from_crossref_api(data)
                pubs.append(publication)
                #print("Fetched: ", doi)
            #pylint: disable-next=W0718
            except Exception as e:
                print(f"Exception for {doi}: ", e)
        session.add_all(pubs)

        print(f"Fetched {len(pubs)}/{len(dois)}")



"""
Entry point for app
"""
from datetime import datetime
from pyjolt import PyJolt, app

from app.configs import Config

@app(__name__, configs=Config)
class App(PyJolt):
    """Main app class"""

    @staticmethod
    def format_datetime(value):
        try:
            dt = datetime.fromisoformat(value)
            return dt.strftime("%Y-%m-%d %H:%M:%S")
        except Exception:
            return value  # fallback if parsing fails



"""
Base model classes for SQLAlchemy models.
"""
import logging
from typing import Any, Type, TypeVar
from sqlalchemy import func
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.future import select
from sqlalchemy.sql import Select

#pylint: disable-next=C0103
_T0 = TypeVar("_T0", bound=Any)

class AsyncQuery:
    """
    Async-friendly intuitive query object.
    Easy and intuitive querying with pagination support.
    """

    def __init__(self, session: AsyncSession, model: Type[_T0]):
        self.session = session
        self.model = model
        self._query: Select = select(model)  # Start with SELECT * FROM table

    def where(self, *conditions) -> "AsyncQuery":
        """Adds WHERE conditions (same as `filter()`)."""
        return self.filter(*conditions)

    def filter(self, *conditions) -> "AsyncQuery":
        """Adds WHERE conditions to the query (supports multiple conditions)."""
        self._query = self._query.filter(*conditions)
        return self

    def filter_by(self, **kwargs) -> "AsyncQuery":
        """Adds WHERE conditions using keyword arguments (simpler syntax)."""
        self._query = self._query.filter_by(**kwargs)
        return self

    def join(self, other_model: Type[_T0]) -> "AsyncQuery":
        """Performs a SQL JOIN with another model."""
        self._query = self._query.join(other_model)
        return self

    def limit(self, num: int) -> "AsyncQuery":
        """Limits the number of results returned."""
        self._query = self._query.limit(num)
        return self

    def offset(self, num: int) -> "AsyncQuery":
        """Skips a certain number of results (used for pagination)."""
        self._query = self._query.offset(num)
        return self

    def order_by(self, *columns) -> "AsyncQuery":
        """Sorts results based on one or more columns."""
        self._query = self._query.order_by(*columns)
        return self
    
    def order_by_strings(self, *args: str) -> "AsyncQuery"  :
        """
        Convert 'column', 'column ASC', or 'column DESC'
        into a SQLAlchemy column expression.
        """
        ordering: list[Any] = []
        for order_str in args:
            parts = order_str.strip().split()

            col_name = parts[0]
            direction = parts[1].upper() if len(parts) > 1 else "ASC"

            # get the model column object
            column = getattr(self.model, col_name)

            if direction == "DESC":
                ordering.append(column.desc())
            ordering.append(column.asc())
        return self.order_by(*ordering)
    
    def like(self, column, pattern, escape=None) -> "AsyncQuery":
        """
        Filters results using a SQL LIKE condition.

        :param column: A column attribute, e.g. User.name
        :param pattern: The pattern string (with % wildcards)
        :param escape: The escape character, if needed.
        :return: self, for chaining
        """
        self._query = self._query.filter(column.like(pattern, escape=escape))
        return self

    def ilike(self, column, pattern, escape=None) -> "AsyncQuery":
        """
        Filters results using a case-insensitive LIKE (ILIKE) condition.
        (Supported by Postgres, some other dialects).

        :param column: A column attribute, e.g. User.name
        :param pattern: The pattern string (with % wildcards)
        :param escape: The escape character, if needed.
        :return: self, for chaining
        """
        self._query = self._query.filter(column.ilike(pattern, escape=escape))
        return self

    async def count(self) -> int:
        """
        Returns the total number of records matching the current query,
        preserving any applied filters—without mutating self._query.
        """
        base = self._query.order_by(None).subquery()
        count_query = select(func.count()).select_from(base)

        try:
            result = await self.session.execute(count_query)
            return int(result.scalar() or 0)
        except SQLAlchemyError as e:
            await self.session.rollback()
            logging.error("Database count query failed: %s", e)
            raise

    async def paginate(self, page: int = 1, per_page: int = 10):
        page = max(page, 1)
        total_records = await self.count()
        total_pages = (total_records + per_page - 1) // per_page

        # IMPORTANT: now safely apply LIMIT/OFFSET to the original SELECT
        q = self._query.limit(per_page).offset((page - 1) * per_page)
        result = await self.session.execute(q)
        items = result.scalars().all()

        return {
            "items": items,
            "total": total_records,
            "page": page,
            "pages": total_pages,
            "per_page": per_page,
            "has_next": page < total_pages,
            "has_prev": page > 1,
        }

    async def _execute_query(self) -> Any:
        """Executes the query safely with automatic rollback on failure."""
        try:
            result = await self.session.execute(self._query)
            return result
        except SQLAlchemyError as e:
            await self.session.rollback()
            logging.error("Database query failed: %s", e)
            raise

    async def all(self) -> list:
        """Executes the query and returns all results."""
        result = await self._execute_query()
        return result.scalars().all()

    async def first(self) -> Any:
        """Executes the query and returns the first result."""
        result = await self._execute_query()
        return result.scalars().first()

    async def one(self) -> Any:
        """Executes the query and expects exactly one result."""
        result = await self._execute_query()
        return result.scalars().one()




"""
migrate.py
Alembic integration for database migrations, with argparse subparsers for CLI commands.
"""

import os
import shutil
from typing import Optional, cast, Any
from alembic.config import Config
from alembic import command
from sqlalchemy import MetaData
from pydantic import BaseModel, Field, ConfigDict

from ....pyjolt import PyJolt
from ..sql_database import SqlDatabase
from ....base_extension import BaseExtension
from ....utilities import to_kebab_case

def register_db_commands(app: PyJolt, migrate: 'Migrate'):
    """
    Registers subparsers (CLI commands) for database migrations to the app's CLI.
    This replaces the old app.add_cli_command(...) calls with subparser definitions
    so that we can parse arguments like "db-upgrade <revision>" or "db-migrate --message '...'",
    etc.
    """

    #init
    sp_init = app.subparsers.add_parser(f"{migrate.command_prefix}-init", help="Initialize the Alembic migration environment.")
    sp_init.set_defaults(func=migrate.init)

    #migrate
    sp_migrate = app.subparsers.add_parser(f"{migrate.command_prefix}-migrate", help="Generate a new revision file (autogenerate).")
    sp_migrate.add_argument("--message", default="", help="Revision message.")
    sp_migrate.set_defaults(func=lambda *args, **kwargs: migrate.migrate(message=kwargs["message"]))

    #upgrade
    sp_upgrade = app.subparsers.add_parser(f"{migrate.command_prefix}-upgrade", help="Upgrade the database to a specified (or head) revision.")
    sp_upgrade.add_argument("--revision", nargs="?", default="head", help="Revision identifier (default=head).")
    sp_upgrade.set_defaults(func=lambda *args, **kwargs: migrate.upgrade(revision=kwargs["revision"]))

    #downgrade
    sp_downgrade = app.subparsers.add_parser(f"{migrate.command_prefix}-downgrade", help="Downgrade the database to a specified (or one step) revision.")
    sp_downgrade.add_argument("--revision", nargs="?", default="-1", help="Revision identifier (default=-1, i.e. one step down).")
    sp_downgrade.set_defaults(func=lambda *args, **kwargs: migrate.downgrade(revision=kwargs["revision"]))

    #history
    sp_history = app.subparsers.add_parser(f"{migrate.command_prefix}-history", help="Show revision history.")
    sp_history.add_argument("--verbose", action="store_true", help="Show more details about each revision.")
    sp_history.add_argument("--indicate-current", action="store_true", help="Mark the current revision in the log.")
    sp_history.set_defaults(func=lambda *args, **kwargs: migrate.history(verbose=kwargs["verbose"],
                                                                         indicate_current=kwargs["indicate_current"]))

    #current
    sp_current = app.subparsers.add_parser(f"{migrate.command_prefix}-current", help="Show the current revision of the database.")
    sp_current.add_argument("--verbose", action="store_true", help="Show more details about the current revision.")
    sp_current.set_defaults(func=lambda *args, **kwargs: migrate.current(verbose=kwargs["verbose"]))

    #heads
    sp_heads = app.subparsers.add_parser(f"{migrate.command_prefix}-heads", help="Show all current 'head' revisions.")
    sp_heads.add_argument("--verbose", action="store_true", help="Show more details.")
    sp_heads.set_defaults(func=lambda *args, **kwargs: migrate.heads(verbose=kwargs["verbose"]))

    #show
    sp_show = app.subparsers.add_parser(f"{migrate.command_prefix}-show", help="Show details of a given revision.")
    sp_show.add_argument("--revision", nargs="?", default="head", help="The revision to show (default=head).")
    sp_show.set_defaults(func=lambda *args, **kwargs: migrate.show(revision=kwargs["revision"]))

    #stamp
    sp_stamp = app.subparsers.add_parser(f"{migrate.command_prefix}-stamp", help="Stamp the database with a given revision (no actual migration).")
    sp_stamp.add_argument("--revision", nargs="?", default="head", help="Revision to stamp (default=head).")
    sp_stamp.set_defaults(func=lambda *args, **kwargs: migrate.stamp(revision=kwargs["revision"]))

class MigrateConfig(BaseModel):
    """Configuration options for Migrate extension"""
    model_config = ConfigDict(extra="allow")

    ALEMBIC_MIGRATION_DIR: str = Field("migrations", description="Connection string for the database")
    ALEMBIC_DATABASE_URI_SYNC: str = Field(description="AsyncSession variable name for use with @managed_session decorator and @readonly_session decorator")

class Migrate(BaseExtension):
    """
    Integrates Alembic with the application for managing database migrations.
    Uses the same configs name as the SqlDatabase instance.

    Commands available in the migration CLI tool are:
    <kebab-case-db-name>-init (default: db-init)
    <kebab-case-db-name>-migrate (default: db-migrate)
    <kebab-case-db-name>-upgrade (default: db-upgrade)
    <kebab-case-db-name>-downgrade
    <kebab-case-db-name>-history
    <kebab-case-db-name>-current
    <kebab-case-db-name>-heads
    <kebab-case-db-name>-show
    <kebab-case-db-name>-stamp
    """
    def __init__(self, db: SqlDatabase):
        self._app: "PyJolt"
        self._root_path: str
        self._db: SqlDatabase = db
        self._configs_name: str = db.configs_name
        self._configs: dict[str, Any] = {}
        self._migrations_path: Optional[str] = None
        self._migration_dir: Optional[str] = None
        self._database_uri: Optional[str] = None
        self._command_prefix: str = to_kebab_case(db.db_name)

    def init_app(self, app: PyJolt):
        """
        Initializes Alembic for the given app.
        """
        self._app = app
        self._root_path = self._app.root_path
        self._configs = app.get_conf(self._configs_name, None)
        if self._configs is None:
            raise ValueError(f"Configurations for {self._configs_name} for {self.__class__.__name__} not found in app configurations.")
        self._configs = self.validate_configs(self._configs, MigrateConfig)
        self._migration_dir = self._configs.get("ALEMBIC_MIGRATION_DIR")
        self._migrations_path = os.path.join(self._root_path, cast(str, self._migration_dir))
        # SYNC driver for migrations
        self._database_uri = self._configs.get("ALEMBIC_DATABASE_URI_SYNC")
        # Register all the subparser commands
        register_db_commands(self._app, self)

    def get_alembic_config(self) -> Config:
        """
        Returns an Alembic configuration object, with target_metadata
        restricted to the models of this SqlDatabase.
        """
        cfg_path = os.path.join(cast(str, self._migrations_path), "alembic.ini")
        config = Config(cfg_path)
        config.set_main_option("sqlalchemy.url", cast(str, self._database_uri))

        associated_models = list(self._db._models.values())
        if not associated_models:
            raise Exception(f"No models associated with db: {self._db.db_name}")

        md = MetaData()
        for model in associated_models:
            model_db_name = getattr(model, "__db_name__", None)
            if model_db_name is not None and model_db_name != self._db.db_name:
                continue
            model.__table__.tometadata(md)

        if not md.tables:
            raise Exception(f"No tables found in metadata for db: {self._db.db_name}")

        config.attributes["target_metadata"] = md
        return config

    def _copy_env_template(self):
        """
        Copies the env.py template to the migrations directory.
        """
        template_path = os.path.join(os.path.dirname(__file__), "env_template.py")
        destination_path = os.path.join(cast(str, self._migrations_path), "env.py")
        shutil.copy(template_path, destination_path)

    def init(self):
        """
        Initializes the Alembic migration environment (if not done already).
        """
        if not os.path.exists(cast(str, self._migrations_path)):
            print("Creating migrations folder...")
            os.makedirs(cast(str, self._migrations_path))
        command.init(self.get_alembic_config(), cast(str, self._migrations_path), template="generic")
        self._copy_env_template()

    def migrate(self, message="Generate migration"):
        """
        Creates a new Alembic revision script (autogenerated) with the given message.
        """
        config = self.get_alembic_config()
        command.revision(config, autogenerate=True, message=message)

    def upgrade(self, revision="head"):
        """
        Upgrades the database to the given revision (default=head).
        """
        config = self.get_alembic_config()
        command.upgrade(config, revision)

    def downgrade(self, revision="-1"):
        """
        Downgrades the database to the given revision (default=-1, i.e. one step).
        """
        config = self.get_alembic_config()
        command.downgrade(config, revision)

    def history(self, verbose=False, indicate_current=False):
        """
        Shows the revision history.
        """
        config = self.get_alembic_config()
        command.history(config, verbose=verbose, indicate_current=indicate_current)

    def current(self, verbose=False):
        """
        Shows the current revision of the database.
        """
        config = self.get_alembic_config()
        command.current(config, verbose=verbose)

    def stamp(self, revision="head"):
        """
        Stamps the database with a given revision, without running migrations.
        """
        config = self.get_alembic_config()
        command.stamp(config, revision)

    def heads(self, verbose=False):
        """
        Displays all current head revisions.
        """
        config = self.get_alembic_config()
        command.heads(config, verbose=verbose)

    def show(self, revision="head"):
        """
        Shows details of a specific revision (default=head).
        """
        config = self.get_alembic_config()
        command.show(config, revision)

    @property
    def command_prefix(self) -> str:
        """
        Returns the command prefix for the CLI
        """
        return self._command_prefix



"""
Logging module
"""
from .logger_config_base import (LoggerBase, LoggerConfig,
                                 LogLevel,
                                 Writable,
                                SinkInput,
                                SinkAccepted,
                                RotationType,
                                RetentionType,
                                CompressionType,
                                FilterType,
                                OutputSink)

__all__ = ["LoggerBase", "LoggerConfig", "LogLevel", "Writable", "SinkInput", "SinkAccepted",
           "RotationType", "RetentionType", "CompressionType", "FilterType",
           "OutputSink"]



"""
Performance file logger
"""
from typing import Any
from pyjolt.logging import LoggerBase

class PerformanceFileLogger(LoggerBase):
    """File logger for performance data"""
    def get_format(self) -> str:
        """Should return a valid format string for the logger output"""
        return ("<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | "
                "<level>{level: <8}</level> | {extra[logger_name]} | "
                "{name}:{function}:{line} - <cyan>{message}</cyan>")

    def get_filter(self):
        def _filter(record: dict[str, Any]) -> bool:
            return "PERFORMANCE" in record["message"]

        return _filter



"""Schemas for researcher model"""
from typing import Optional, TYPE_CHECKING, Self
from pydantic import BaseModel

from .base_schema import BaseResponse

if TYPE_CHECKING:
    from app.api.models.researcher import Researcher

class ResearcherOutSchema(BaseModel):
    """Researcher output schema"""
    id: int
    fullname: str
    code: str
    email: str
    education: Optional[str] = None
    employment: Optional[str] = None
    title: Optional[str] = None

    @classmethod
    def from_model(cls, researcher: "Researcher") -> Self:
        """Creates a schema from model"""
        return cls(
            id=researcher.id,
            fullname=researcher.fullname,
            code=researcher.code,
            email=researcher.email,
            education=researcher.education,
            employment=researcher.employment,
            title=researcher.title
        )

class ResearcherResponseSchema(BaseResponse):
    """Researcher response schema"""
    data: ResearcherOutSchema

class AllResearchersResponseSchema(BaseResponse):
    """All researchers out schema"""
    data: list[ResearcherOutSchema]



"""

Revision ID: 0bc5d9cff5b5
Revises: 
Create Date: 2025-12-08 08:35:26.685448

"""
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = '0bc5d9cff5b5'
down_revision: Union[str, None] = None
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('dummies',
    sa.Column('fullname', sa.Text(), nullable=False),
    sa.Column('email', sa.Text(), nullable=False),
    sa.Column('title', sa.Text(), nullable=True),
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('created_at', sa.DateTime(timezone=True), nullable=False),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('email')
    )
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('dummies')
    # ### end Alembic commands ###



"""
NoSQL Database Module
"""
from pydantic import BaseModel, Field
from typing import (Optional, Callable, Any,
                    Iterable, Mapping, TYPE_CHECKING,
                    TypedDict, cast, Type, NotRequired)
from functools import wraps

from .backends.async_nosql_backend_protocol import AsyncNoSqlBackendBase
from ...base_extension import BaseExtension
from ...utilities import run_sync_or_async


if TYPE_CHECKING:
    from ...pyjolt import PyJolt

class _NoSqlDatabaseConfig(BaseModel):
    """Configuration options for NoSqlDatabase extension."""
    BACKEND: Type[AsyncNoSqlBackendBase] = Field(description='Backend class implementing AsyncNoSqlBackendBase.')
    DATABASE_URI: str = Field(description="Connection string for the NoSQL backend.")
    DATABASE_NAME: Optional[str] = Field(default=None, description="Database / keyspace name (if backend uses it).")
    DB_INJECT_NAME: Optional[str] = Field(default="db", description="Kwarg name injected by decorators (database handle).")
    SESSION_NAME: Optional[str] = Field(default="session", description="Kwarg name injected by @managed_database (session/transaction handle).")

class NoSqlDatabaseConfig(TypedDict):
    BACKEND: Type[AsyncNoSqlBackendBase]
    DATABASE_URI: str
    DATABASE_NAME: NotRequired[str]
    INJECT_NAME: NotRequired[str]
    SESSION_NAME: NotRequired[str]

class NoSqlDatabase(BaseExtension):
    """
    A simple async NoSQL Database interface with pluggable backends.
    """

    def __init__(self, db_name: str = "nosql", configs_name: str = "NOSQL_DATABASE") -> None:
        self._app: Optional["PyJolt"] = None
        self._configs_name = configs_name
        self._configs: dict[str, Any] = {}
        self.__db_name__ = db_name

        # Effective config
        self.backend_name: Optional[str] = None
        self.uri: Optional[str] = None
        self.database: Optional[str] = None
        self.inject_name: str = "db"
        self.session_name: str = "session"

        # Backend instance
        self._backend: AsyncNoSqlBackendBase = cast(AsyncNoSqlBackendBase, None)

    # ---- App lifecycle ----

    def init_app(self, app: "PyJolt") -> None:
        """
        Initializes the NoSQL interface.
        Required config keys (with optional variable_prefix):
            - BACKEND
            - DATABASE_URI
        Optional:
            - DATABASE
            - INJECT_NAME (default "db")
            - SESSION_NAME (default "session")
        """
        self._app = app
        self._configs = app.get_conf(self._configs_name, None)
        if self._configs is None:
            raise ValueError(f"Missing {self._configs_name} configuration.")
        self._configs = self.validate_configs(self._configs, _NoSqlDatabaseConfig)

        self.backend_cls = self._configs.get("BACKEND")
        self.uri = self._configs.get("DATABASE_URI")
        self.database = self._configs.get("DATABASE_NAME")
        self.inject_name = cast(str, self._configs.get("INJECT_NAME"))
        self.session_name = cast(str, self._configs.get("SESSION_NAME"))

        if not self.backend_cls or not self.uri:
            raise RuntimeError("Missing NOSQL_BACKEND or NOSQL_DATABASE_URI configuration.")
        if not issubclass(self.backend_cls, AsyncNoSqlBackendBase):
            raise RuntimeError("NOSQL_BACKEND must be a subclass of AsyncNoSqlBackendBase.")

        self._backend = cast(AsyncNoSqlBackendBase, self.backend_cls).configure_from_app(app, self._configs)

        app.add_extension(self)
        app.add_on_startup_method(self.connect)
        app.add_on_shutdown_method(self.disconnect)

    async def connect(self) -> None:
        """
        Creates the backend client/connection. Runs on lifespan.start
        """
        if not self._backend:
            raise RuntimeError("Backend not initialized. Call init_app() first.")
        await self._backend.connect()

    async def disconnect(self) -> None:
        """
        Disposes backend resources. Runs on lifespan.shutdown
        """
        if self._backend:
            await self._backend.disconnect()

    @property
    def db_name(self) -> str:
        return self.__db_name__

    @property
    def backend(self) -> AsyncNoSqlBackendBase:
        if not self._backend:
            raise RuntimeError("Backend not connected. Was init_app/connect called?")
        return self._backend

    def database_handle(self) -> Any:
        return self.backend.database_handle()

    def get_collection(self, name: str) -> Any:
        return self.backend.get_collection(name)

    async def find_one(self, collection: str, filter: Mapping[str, Any], **kwargs) -> Any:
        return await self.backend.find_one(collection, filter, **kwargs)

    async def find_many(self, collection: str, filter: Mapping[str, Any] | None = None, **kwargs) -> list[Any]:
        return await self.backend.find_many(collection, filter, **kwargs)

    async def insert_one(self, collection: str, doc: Mapping[str, Any], **kwargs) -> Any:
        return await self.backend.insert_one(collection, doc, **kwargs)

    async def insert_many(self, collection: str, docs: Iterable[Mapping[str, Any]], **kwargs) -> Any:
        return await self.backend.insert_many(collection, docs, **kwargs)

    async def update_one(self, collection: str, filter: Mapping[str, Any], update: Mapping[str, Any], **kwargs) -> Any:
        return await self.backend.update_one(collection, filter, update, **kwargs)

    async def delete_one(self, collection: str, filter: Mapping[str, Any], **kwargs) -> Any:
        return await self.backend.delete_one(collection, filter, **kwargs)

    async def aggregate(self, collection: str, pipeline: Iterable[Mapping[str, Any]], **kwargs) -> list[Any]:
        return await self.backend.aggregate(collection, pipeline, **kwargs)

    async def execute_raw(self, *args, **kwargs) -> Any:
        """
        Escape hatch for backend-specific commands. See MongoBackend.execute_raw docstring.
        """
        return await self.backend.execute_raw(*args, **kwargs)


    @property
    def managed_database(self) -> Callable:
        """
        Decorator that:
        - Injects a database/client handle into handler kwargs under NOSQL_INJECT_NAME ("db" by default).
        - If the backend supports transactions, opens a session+transaction for the duration
          and commits/aborts automatically on exit.
        """

        def decorator(handler: Callable) -> Callable:
            @wraps(handler)
            async def wrapper(*args, **kwargs):
                backend = self.backend
                inject_key = self.inject_name
                handle = backend.database_handle()

                # default injection
                kwargs[inject_key] = handle

                if not backend.supports_transactions():
                    # No transaction semantics; just run.
                    return await run_sync_or_async(handler, *args, **kwargs)

                # Transaction-capable flow
                session = await backend.start_session()

                async def _call_with_session(*_args, **_kwargs):
                    # Some backends (Mongo) accept 'session' in calls; handlers can pass it through.
                    _kwargs.setdefault(self.session_name, session)
                    return await run_sync_or_async(handler, *_args, **_kwargs)

                try:
                    return await backend.with_transaction(_call_with_session, *args, **kwargs, session=session)
                finally:
                    #session ended in with_transaction, but just in case
                    try:
                        await session.end_session()  # type: ignore[attr-defined]
                    except Exception:
                        pass

            return wrapper

        return decorator



"""
Task manager
"""
import asyncio
from pyjolt.task_manager import TaskManager, schedule_job

class SchedulerManager(TaskManager):
    """
    Custom TaskManager implementation
    """
    @schedule_job("interval", minutes=1, id="some_task", name="A simple task")
    async def some_task(self):
        self.app.logger.info("Started job...")
        await asyncio.sleep(3)
        self.app.logger.info("Finished job...")

scheduler_manager: SchedulerManager = SchedulerManager()



"""
Base logger configuration class
"""
import sys
import re
from abc import ABC
from datetime import timedelta
from pathlib import Path
from typing import (Dict, Any, Union, Optional, ClassVar, cast, TYPE_CHECKING,
                    IO, Callable, Protocol, runtime_checkable, Type, TypedDict,
                    NotRequired)
from enum import StrEnum
from pydantic import BaseModel, field_validator, Field, ConfigDict

from loguru import logger

if TYPE_CHECKING:
    from ..pyjolt import PyJolt

@runtime_checkable
#pylint: disable=R0903
class Writable(Protocol):
    def write(self, s: str) -> Any: ...

SinkType = Union[str, Path, IO[str], IO[bytes], Callable[[str], Any]]
SinkInput = Union[str, Path, IO[str], IO[bytes], Callable[[str], Any], Writable]
SinkAccepted = Union[str, Writable, Callable[[str], None]]
RotationType = Union[str, int, timedelta, Callable[[str, Any], Any]]
RetentionType = Union[str, int, timedelta, Callable[[str, Any], Any]]
CompressionType = Optional[str]
FilterType = Union[None, str, Dict[str, str], Callable[[Dict[str, Any]], bool]]

class OutputSink(StrEnum):
    STDERR = "STDERR"
    STDOUT = "STDOUT"

class LogLevel(StrEnum):
    TRACE = "TRACE"
    DEBUG = "DEBUG"
    INFO = "INFO"
    WARNING = "WARNING"
    ERROR = "ERROR"
    CRITICAL = "CRITICAL"

class _LoggerConfigBase(BaseModel):
    """
    Base config model for loggers
    """
    model_config = ConfigDict(extra="allow")

    SINK: Optional[str|Path] = Field(OutputSink.STDERR.value, description="Output of the logger")
    LEVEL: Optional[LogLevel] = Field(LogLevel.TRACE, description="Log level for logger implementation")
    FORMAT: Optional[str] = Field("<green>{time:HH:mm:ss}</green> | <level>{level}</level> | {extra[logger_name]} | <level>{message}</level>", description="Output format")
    ENQUEUE: Optional[bool] = Field(False, description="Use enqueue or not. For thread, process and coroutine safe logging.")
    BACKTRACE: Optional[bool] = Field(True, description="Use backtrace")
    DIAGNOSE: Optional[bool] = Field(True, description="Diagnose mode")
    COLORIZE: Optional[bool] = Field(True, description="Colorize output")
    DELAY: Optional[bool] = Field(True, description="Use delay")
    ROTATION: Optional[RotationType] = Field("5 MB", description="Size of log file before rotating.")
    RETENTION: Optional[RetentionType] = Field("10 days", description="How long log files are kept before automatic erasure.")
    COMPRESSION: CompressionType = Field("zip", description="If log files should be compressed. No compression = None")
    SERIALIZE: Optional[bool] = Field(False)
    ENCODING: Optional[str] = Field("utf-8", description="Encoding of logged messages - for file logging.")
    MODE: Optional[str] = Field("a", description="If log messages should be appended to the file.")

    @field_validator("SINK", mode="before")
    @classmethod
    def normalize_sink(cls, v):
        """Accept case-insensitive strings and convert to canonical form."""
        s = v.strip().upper()
        if s in {"STDERR", "SYS.ERR", "CONSOLE", "ERR"}:
            return "STDERR"
        if s in {"STDOUT", "SYS.OUT", "OUT"}:
            return "STDOUT"
        if s in {"NULL", "DEVNULL", "NONE"}:
            return "NULL"
        # Otherwise treat it as a path
        return Path(v)

class LoggerConfig(TypedDict):
    SINK: NotRequired[str|Path]
    LEVEL: NotRequired[LogLevel]
    FORMAT: NotRequired[str]
    ENQUEUE: NotRequired[bool]
    BACKTRACE: NotRequired[bool]
    DIAGNOSE: NotRequired[bool]
    COLORIZE: NotRequired[bool]
    DELAY: NotRequired[bool]
    ROTATION: NotRequired[RotationType]
    RETENTION: NotRequired[RetentionType]
    COMPRESSION: NotRequired[CompressionType]
    SERIALIZE: NotRequired[bool]
    ENCODING: NotRequired[str]
    MODE: NotRequired[str]

class LoggerBase(ABC):
    """
    Base template for configuring a Loguru sink.
    Subclasses implement get_sink() and optionally override getters below.
    """
    configs_model: ClassVar[Optional[Type[_LoggerConfigBase]]] = cast(Type[_LoggerConfigBase], None)

    def __init__(self, app: "PyJolt"):
        self.app: "PyJolt" = app
        #loads configs for the logger from application configurations
        #by the config class name as upper-case 
        #example: CustomLoggerConfig -> CUSTOM_LOGGER_CONFIG
        self.conf: Dict[str, Any] = app.get_conf(self.logger_name, None) or {} # type: ignore
        self.conf = self.validate_configs(self.conf)

    def validate_configs(self, configs: dict[str, Any]) -> dict[str, Any]:
        if self.configs_model is not None and issubclass(self.configs_model, _LoggerConfigBase):
            return self.configs_model.model_validate(configs).model_dump()
        return _LoggerConfigBase.model_validate(configs).model_dump()

    @property
    def logger_name(self) -> str:
        """Returns class name as upper snake case"""
        name = self.__class__.__name__
        return re.sub(r'(?<!^)(?=[A-Z])', '_', name).upper()

    def get_sink(self) -> SinkInput:
        sink = self.conf.get("SINK", "STDERR")
        if sink == "STDERR":
            return sys.stderr
        if sink == "STDOUT":
            return sys.stdout
        if sink == "NULL":
            return "NUL" if sys.platform.startswith("win") else "/dev/null"
        p = Path(sink)
        p.parent.mkdir(parents=True, exist_ok=True)
        return p

    def get_conf_value(self, key: str, default: Any = None) -> Any:
        return self.conf.get(key, default)

    def get_level(self) -> Union[str, int]:
        return self.get_conf_value("LEVEL", LogLevel.INFO)

    def get_format(self) -> str:
        # includes the logger name as a constant-like tag
        return self.get_conf_value(
            "FORMAT",
            "<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | "
            "<level>{level: <8}</level> | {extra[logger_name]} | "
            "{name}:{function}:{line} - <cyan>{message}</cyan>",
        )

    def dict_to_format_string(self, fmt_dict: dict[str, str]) -> str:
        """
        Converts a dictionary of Loguru placeholders into a JSON-like
        format string suitable for loguru.add(format=...).

        CURRENTLY DOESN'T WORK AS INTENDED
        """
        items = []
        for key, value in fmt_dict.items():
            items.append(f'"{key}": "{value}"')
        return "{{{}}}".format(", ".join(items))

    def get_rotation(self) -> Optional[RotationType]:
        return self.get_conf_value("ROTATION", None)

    def get_retention(self) -> Optional[RetentionType]:
        return self.get_conf_value("RETENTION", None)

    def get_compression(self) -> CompressionType:
        return self.get_conf_value("COMPRESSION", None)

    def get_filter(self) -> FilterType:
        return None

    def _wrap_filter_with_logger_name(self,
            original_filter: FilterType) -> Callable[[Dict[str, Any]], bool]:
        def _wrapped(record: Dict[str, Any]) -> bool:
            # Injects the logger name as extra information
            record["extra"].setdefault("logger_name", self.logger_name)

            if original_filter is None:
                return True
            if callable(original_filter):
                return bool(original_filter(record))
            if isinstance(original_filter, str):
                return record.get("name") == original_filter
            if isinstance(original_filter, dict):
                return all(record["extra"].get(k) == v for k, v in original_filter.items())
            return True
        return _wrapped

    def get_enqueue(self) -> bool:
        return self.get_conf_value("ENQUEUE", True)

    def get_backtrace(self) -> bool:
        return self.get_conf_value("BACKTRACE", False)

    def get_diagnose(self) -> bool:
        return self.get_conf_value("DIAGNOSE", False)

    def get_colorize(self) -> Optional[bool]:
        return self.get_conf_value("COLORIZE", None)

    def get_serialize(self) -> bool:
        return self.get_conf_value("SERIALIZE", False)

    def get_encoding(self) -> Optional[str]:
        return self.get_conf_value("ENCODING", "utf-8")

    def get_mode(self) -> Optional[str]:
        return self.get_conf_value("MODE", "a")

    def get_delay(self) -> bool:
        return self.get_conf_value("DELAY", True)

    def remove_existing_handlers(self) -> bool:
        return self.get_conf_value("remove_existing_handlers", False)

    def _build_handler_kwargs(self) -> Dict[str, Any]:
        return {
            "level": self.get_level(),
            "format": self.get_format(),
            "rotation": self.get_rotation(),
            "retention": self.get_retention(),
            "compression": self.get_compression(),
            "filter": self.get_filter(),
            "enqueue": self.get_enqueue(),
            "backtrace": self.get_backtrace(),
            "diagnose": self.get_diagnose(),
            "colorize": self.get_colorize(),
            "serialize": self.get_serialize(),
            "encoding": self.get_encoding(),
            "mode": self.get_mode(),
            "delay": self.get_delay(),
        }

    def _is_path_sink(self, sink: SinkAccepted) -> bool:
        return isinstance(sink, str)

    def _is_stream_sink(self, sink: SinkAccepted) -> bool:
        return (not isinstance(sink, str)) and hasattr(sink, "write")

    def _is_callable_sink(self, sink: SinkAccepted) -> bool:
        return callable(sink) and not hasattr(sink, "write")

    def _normalize_sink(self, sink: SinkInput) -> SinkAccepted:
        if isinstance(sink, Path):
            return sink.as_posix()
        return sink  # type: ignore[return-value]

    def _filter_kwargs_for_sink(self, sink: SinkAccepted, kwargs: Dict[str, Any]) -> Dict[str, Any]:
        # Common kwargs accepted by all sinks
        common = {
            "level", "format", "filter",
            "colorize", "serialize", "backtrace", "diagnose",
            "enqueue", "catch"
        }
        # File-path only kwargs (Loguru opens the file for you)
        file_only = {"rotation", "retention", "compression",
                     "mode", "delay", "encoding", "buffering"}
        # Streams (TextIO / Writable): no rotation/retention/compression/encoding/mode/delay
        stream_only: set = set()  # (no extra)
        # Callables: also no file-only kwargs
        callable_only: set = set()

        if self._is_path_sink(sink):
            allowed = common | file_only
        elif self._is_stream_sink(sink):
            allowed = common | stream_only
        else:  # callable sink
            allowed = common | callable_only

        return {k: v for k, v in kwargs.items() if k in allowed and v is not None}

    def configure(self) -> int:

        raw_sink = self.get_sink()
        sink = self._normalize_sink(raw_sink)

        base_kwargs = self._build_handler_kwargs()
        # Wrap filter to inject extra["logger_name"] for every record to this sink
        base_kwargs["filter"] = self._wrap_filter_with_logger_name(base_kwargs.get("filter"))

        kwargs = self._filter_kwargs_for_sink(sink, base_kwargs)
        logger_sink_id: int = logger.add(sink, **kwargs)

        return logger_sink_id



"""
Redis cache implementation

CACHE_REDIS_URL       = "redis://localhost:6379/0"   # required
CACHE_REDIS_PASSWORD  = None                          # optional
CACHE_DURATION        = 300                           # optional (default TTL)
CACHE_KEY_PREFIX      = "pyjolt:cache:"              # optional prefix/namespace
"""
from __future__ import annotations

import pickle
from typing import Optional, cast, List, TYPE_CHECKING, Any

from redis.asyncio import Redis, from_url

from .base_cache_backend import BaseCacheBackend

if TYPE_CHECKING:
    from ...pyjolt import PyJolt

class RedisCacheBackend(BaseCacheBackend):
    """Redis-backed cache using binary pickled payloads."""

    def __init__(
        self,
        url: str,
        password: Optional[str] = None,
        default_ttl: int = 300,
        key_prefix: str = ""
    ) -> None:
        if not url:
            raise ValueError("CACHE_REDIS_URL must be set for RedisCacheBackend")
        self._url = url
        self._password = password
        self._client: Optional[Redis] = None
        self.default_ttl = int(default_ttl)
        # Normalize prefix to empty or end with ':' for nicer namespaces
        if key_prefix and not key_prefix.endswith(":") and key_prefix != "":
            key_prefix = key_prefix + ":"
        self._prefix = key_prefix

    @classmethod
    def configure_from_app(cls, app: PyJolt, configs: dict[str, Any]) -> "RedisCacheBackend":
        url = configs.get("REDIS_URL", "")
        password = configs.get("REDIS_PASSWORD", None)
        ttl = cast(int, configs.get("DURATION"))
        key_prefix = configs.get("KEY_PREFIX", "")
        return cls(url=url, password=password, default_ttl=ttl, key_prefix=key_prefix)

    async def connect(self) -> None:
        if not self._client:
            # decode_responses=False -> bytes in/out, good for pickled values
            self._client = await from_url(
                self._url,
                encoding="utf-8",
                decode_responses=False,
                password=self._password,
            )

    async def disconnect(self) -> None:
        if self._client:
            await self._client.close()
            self._client = None

    # ---- helpers ----
    def _k(self, key: str) -> str:
        return f"{self._prefix}{key}" if self._prefix else key

    async def _ensure(self) -> Redis:
        if not self._client:
            # Allow lazy connect if caller forgot to call connect()
            await self.connect()
        assert self._client is not None
        return self._client

    async def get(self, key: str) -> Optional[dict]:
        client = await self._ensure()
        raw = await client.get(self._k(key))
        if not raw:
            return None
        return cast(dict, pickle.loads(raw))

    async def set(self, key: str, value: dict, duration: Optional[int] = None) -> None:
        client = await self._ensure()
        ttl = int(duration) if duration is not None else self.default_ttl
        await client.setex(self._k(key), ttl, pickle.dumps(value))

    async def delete(self, key: str) -> None:
        client = await self._ensure()
        await client.delete(self._k(key))

    async def clear(self) -> None:
        client = await self._ensure()
        if self._prefix and self._prefix != "":
            keys: List[bytes] = []
            async for k in client.scan_iter(match=f"{self._prefix}*"):
                keys.append(k)
            if keys:
                # Pipeline deletion in chunks to avoid huge payloads
                #pylint: disable-next=C0103
                CHUNK = 1000
                for i in range(0, len(keys), CHUNK):
                    chunk = keys[i : i + CHUNK]
                    await client.delete(*chunk)
        else:
            #flushes entire Redis db in case no namespace prefix is configured
            await client.flushdb()



"""
Admin dashboard email clients controller.
Handles email client extension for sending emails.
"""
# pylint: disable=W0719,W0212
from __future__ import annotations
from typing import Optional, cast
from pydantic import BaseModel, ValidationError, EmailStr

from .common_controller import CommonAdminController
from ..email.email_client import EmailClient
from ..auth.authentication import login_required
from ..controller import get, post
from ..http_statuses import HttpStatus
from ..request import Request
from ..response import Response
from ..utilities import base64_to_bytes, to_kebab_case


class EmailQueryParam(BaseModel):
    """Email validation"""
    client: str
    query: Optional[str] = None

class SendEmailModel(BaseModel):
    """Model for validating emails to be sent"""
    to_address: list[EmailStr]
    subject: str
    attachments: Optional[dict[str, str]] = None
    body: str

class AdminEmailClientsController(CommonAdminController):
    """Admin dashboard controller."""

    @get("/email-clients")
    @login_required
    async def email_clients(self, req: Request) -> Response:
        """Email clients page"""

        return await req.res.html(
            "/__admin_templates/email_clients.html", {
                "email_clients": self.dashboard.email_clients,
                **self.get_common_variables()
            }
        )

    @get("/send-email")
    @login_required
    async def send_email(self, req: Request) -> Response:
        """
        Endpoint for sending emails
        """
        client_query = EmailQueryParam.model_validate(req.query_params)
        client = self.get_email_client(client_query.client)
        return await req.res.html("/__admin_templates/send_email.html", {
            "client_name": to_kebab_case(client.configs_name),
            "client": client, **self.get_common_variables()
        })

    @get("/email-query")
    @login_required
    async def email_query(self, req: Request) -> Response:
        """Endpoint for email querying"""
        client_query = EmailQueryParam.model_validate(req.query_params)
        client = self.get_email_client(client_query.client)
        results: list[tuple[str, str]] = await self.dashboard.email_recipient_query(req, 
                                                cast(str, client_query.query), client)
        return req.res.json({
            "message": "Email query results",
            "status": "success",
            "data": results
        }).status(HttpStatus.OK)
    
    @post("/email-submit")
    @login_required
    async def email_submit(self, req: Request) -> Response:
        """Sends the email"""
        try:
            client_query = EmailQueryParam.model_validate(req.query_params)
            client = self.get_email_client(client_query.client)
            json_data: dict = cast(dict, await req.json())
            message = SendEmailModel.model_validate(json_data)
            attachments: Optional[dict[str, bytes]] = None
            if message.attachments is not None:
                attachments = {}
                for key, value in message.attachments.items():
                    attachments[key] = base64_to_bytes(value)

            await client.send_email(
                to_address=message.to_address,
                subject=message.subject,
                body=message.body,
                attachments=attachments
            )
            self.app.logger.info(f"Email sent with client {client.configs_name} to addresses: {message.to_address} with subject {message.subject}")
            return req.res.json({
                "message": "Email sent successfully",
                "status": "success",
            }).status(HttpStatus.OK)
        except ValidationError as exc:
            details = {}
            if hasattr(exc, "errors"):
                for error in exc.errors():
                    details[error["loc"][0]] = error["msg"]
            return req.response.json({
                "message": "Missing data.",
                "details": details
            }).status(HttpStatus.UNPROCESSABLE_ENTITY)
        except Exception as exc:
            self.app.logger.debug(exc)
            return req.res.json({
                "message": "Something went wrong",
                "status": "error",
            }).status(HttpStatus.INTERNAL_SERVER_ERROR)
 
    def get_email_client(self, client_name: str) -> EmailClient:
        client = None
        if self.dashboard.email_clients is None:
            raise Exception("No registered email clients.")
        for name, email_client in self.dashboard.email_clients.items():
            if name == client_name:
                client = email_client
                break
        if client is None:
            raise Exception("Unknown email client.")
        return cast(EmailClient, client)




"""
sql_database.py
Module for sql database connection/integration
"""

#import asyncio
from typing import (Any, Dict, Optional,
                    Callable, Tuple, Type,
                    TypedDict, cast, TYPE_CHECKING,
                    NotRequired)
from functools import wraps
from sqlalchemy import MetaData, Table, select, func
from sqlalchemy.inspection import inspect
from sqlalchemy.engine import RowMapping
from sqlalchemy.ext.asyncio import (
    AsyncEngine,
    create_async_engine,
    AsyncSession,
    async_sessionmaker
)
from pydantic import BaseModel, Field, ConfigDict
from .dialect_overview_extras import (_portable_schema_stats,
                                        _extras_postgres,
                                        _extras_sqlite,
                                        _extras_mysql,
                                        _extras_mssql,
                                        _extras_oracle)
#pylint: disable-next=E0402
from ...utilities import run_sync_or_async
#pylint: disable-next=E0402
from ...base_extension import BaseExtension
from .declarative_base import DeclarativeBaseModel
if TYPE_CHECKING:
    from ...pyjolt import PyJolt

class _SqlDatabaseConfig(BaseModel):
    """Configuration options for SqlDatabase extension"""
    model_config = ConfigDict(extra="allow")

    DATABASE_URI: str = Field(description="Connection string for the database")
    DATABASE_SESSION_NAME: str = Field("session",
                                description=("AsyncSession variable name for use "
                                            "with @managed_session decorator and "
                                            "@readonly_session decorator"))
    SHOW_SQL: bool = Field(False,
        description="If True, every executed SQL statement is logged to the console.")
    NICE_NAME: Optional[str] = Field(None, description="Name of the database for admin dashboard view. Defaults to db_name variable.")

class SqlDatabaseConfig(TypedDict):
    DATABASE_URI: str
    DATABASE_SESSION_NAME: NotRequired[str]
    SHOW_SQL: bool
    NICE_NAME: NotRequired[str]
    ALEMBIC_DATABASE_URI_SYNC: NotRequired[str]
    ALEMBIC_MIGRATION_DIR: NotRequired[str]

_DIALECT_EXTRAS: Dict[str, Callable] = {
    "postgresql": _extras_postgres,
    "sqlite": _extras_sqlite,
    "mysql": _extras_mysql,
    "mariadb": _extras_mysql,
    "mssql": _extras_mssql,
    "oracle": _extras_oracle,
}

class SqlDatabase(BaseExtension):
    """
    A simple async Database interface using SQLAlchemy.
    """

    def __init__(self, db_name: Optional[str] = None, configs_name: str = "SQL_DATABASE") -> None:
        self._app: "Optional[PyJolt]" = None
        self._engine: Optional[AsyncEngine] = None
        self._session_factory: Optional[async_sessionmaker[AsyncSession]] = None
        self._db_uri: str = ""
        self._configs_name: str = cast(str, configs_name)
        self._configs: dict[str, str] = {}
        if db_name is None:
            db_name = configs_name.lower()
        self.__db_name__ = db_name
        self._session_name: str
        self._models: dict[str, type[DeclarativeBaseModel]] = {}
        self._number_of_tables: int = 0

    def init_app(self, app: "PyJolt") -> None:
        """
        Initializes the database interface
        app.get_conf("DATABASE_URI") must return a connection string like:
        "postgresql+asyncpg://user:pass@localhost/dbname"
        or "sqlite+aiosqlite:///./pyjolt.db"
        """
        self._app = app
        self._configs = app.get_conf(self._configs_name, None)
        if self._configs is None:
            raise ValueError(f"Configurations for {self._configs_name} not found in app configurations.")
        self._configs = self.validate_configs(self._configs, _SqlDatabaseConfig)
        self._db_uri = self._configs["DATABASE_URI"]
        self._session_name = self._configs["DATABASE_SESSION_NAME"]
        self._app.add_extension(self)
        self._app.add_on_startup_method(self.connect)
        self._app.add_on_shutdown_method(self.disconnect)
        for model in self._app._db_models.get(self.__db_name__, []):
            self._models[model.__name__] = model

    async def connect(self) -> None:
        """
        Creates the async engine and session factory.
        Runs automatically when the lifespan.start signal is received
        """
        if not self._engine:
            self._engine = create_async_engine(
                cast(str, self._db_uri),
                echo=cast(bool, self._configs.get("SHOW_SQL", False)),
                pool_pre_ping=True,
                pool_recycle=1800
            )

        self._session_factory = async_sessionmaker(
            bind=self._engine,
            expire_on_commit=False,
            autoflush=False
        )
    
    async def disconnect(self) -> None:
        """
        Runs automatically when the lifespan.shutdown signal is received
        """
        if self._engine:
            await self._engine.dispose()
            self._engine = None

    def create_session(self) -> AsyncSession:
        """
        Creates new session and returns session object. Used for manual session handling.

        WARNING: You must close the session manually after use with await session.close()
        or use it within an async with block.
        """
        if self._session_factory is not None:
            return cast(AsyncSession, self._session_factory())
        #pylint: disable-next=W0719
        raise Exception("Session factory is None")

    async def execute_raw(self, statement, *, as_transaction: bool = False) -> list[RowMapping]:
        """
        Executes raw sql statement and returns list of RowMapping objects.
        
        If as_transaction is True, the execution will be wrapped in a transaction.
        as_transaction=False is for read-only state,emts; DML 
        """
        if not self._session_factory:
            raise RuntimeError("Database is not connected.")
        async with self._session_factory() as session:
            if as_transaction:
                async with session.begin():
                    result = await session.execute(statement)
            else:
                result = await session.execute(statement)
            return cast(list[RowMapping],result.mappings().all())
    
    async def count_tables(self, schema: str | None = None) -> int:
        if self._engine is None:
            return 0
        async with self._engine.connect() as conn:
            return await conn.run_sync(
                lambda sync_conn: len(inspect(sync_conn).get_table_names(schema=schema))
            )
        
    async def count_rows_exact(self, schema: str | None = None
    ) -> Tuple[Dict[str, int], int]:
        """
        Returns (per_table_counts, total_rows). Counts are exact but can be slow.
        """
        if self._engine is None:
            raise Exception("Missing engine for database: ", self.__class__.__name__)
        async with self._engine.connect() as aconn:
            def _work(sync_conn) -> Tuple[Dict[str, int], int]:
                insp = inspect(sync_conn)
                table_names = insp.get_table_names(schema=schema)

                md = MetaData()
                per_table: Dict[str, int] = {}
                total = 0

                for name in table_names:
                    t = Table(name, md, schema=schema, autoload_with=sync_conn)
                    cnt = sync_conn.execute(
                        select(func.count()).select_from(t)
                    ).scalar_one()
                    full = f"{schema}.{name}" if schema else name
                    per_table[full] = int(cnt)
                    total += int(cnt)

                return per_table, total

            return await aconn.run_sync(_work)
    
    async def collect_db_overview(self, schema: str | None = None, with_extras: bool = True) -> Dict[str, Any]:
        """
        Portable overview across dialects, with optional per-dialect extras.
        """
        if self._engine is None:
            raise Exception("Missing engine for database: ", self.__class__.__name__)
        async with self._engine.connect() as aconn:
            def _work(sync_conn):
                data = _portable_schema_stats(sync_conn, schema=schema)
                if with_extras:
                    dname = sync_conn.dialect.name
                    extras_fn = _DIALECT_EXTRAS.get(dname)
                    if extras_fn:
                        try:
                            data["extras"] = extras_fn(sync_conn)
                        except Exception:
                            # Don't break the dashboard if extras fail
                            data["extras"] = {"error": f"{dname} extras unavailable"}
                return data

            return await aconn.run_sync(_work)

    @property
    def db_uri(self):
        """
        Returns database connection uri string
        """
        return self._db_uri

    @property
    def engine(self) -> AsyncEngine:
        """
        Returns database engine
        """
        if self._engine is None:
            raise RuntimeError("Engine not initialized. Call connect() first.")
        return cast(AsyncEngine, self._engine)
    
    @property
    def session_name(self) -> str:
        """
        Returns the session variable name to be used in the kwargs of the request handler.
        Default is "session", can be changed via configuration.
        """
        return self._session_name

    @property
    def db_name(self) -> str:
        return self.__db_name__

    @property
    def nice_name(self) -> str:
        name: str = cast(str, self._configs.get("NICE_NAME"))
        if name is None:
            return self.db_name
        return name
    
    @property
    def models_list(self) -> list[Type[DeclarativeBaseModel]]:
        """List fo all models"""
        return [model for model in self._models.values()]

    @property
    def managed_session(self) -> Callable:
        """
        Returns a decorator that:
        - Creates a new AsyncSession per request.
        - Injects it into the kwargs of the request with the key "session" or custom session name.
        - Commits if no error occurs.
        - Rolls back if an unhandled error occurs.
        - Closes the session automatically afterward.
        """

        def decorator(handler: Callable) -> Callable:
            @wraps(handler)
            async def wrapper(*args, **kwargs):
                if not self._session_factory:
                    raise RuntimeError(
                        "Database is not connected. "
                        "Connection should be established automatically."
                        "Please check network connection and configurations."
                    )
                async with self._session_factory() as session:  # Ensures session closure
                    async with session.begin():  # Ensures transaction handling (auto commit/rollback)
                        kwargs[self.session_name] = session
                        return await run_sync_or_async(handler, *args, **kwargs)
            return wrapper
        return decorator
    
    @property
    def managed_session_for_cli(self) -> Callable:
        """
        A managed session for CLI commands which first connects to the DB
        and handles session injection and connection closing
        - Creates a new AsyncSession per request.
        - Injects it into the kwargs of the request with the key "session" or custom session name.
        - Commits if no error occurs.
        - Rolls back if an unhandled error occurs.
        - Closes the session automatically afterward.
        """

        def decorator(handler: Callable) -> Callable:
            @wraps(handler)
            async def wrapper(*args, **kwargs):
                await self.connect() #connects to db
                assert self._session_factory is not None
                async with self._session_factory() as session:  # Ensures session closure
                    async with session.begin():  # Ensures transaction handling (auto commit/rollback)
                        kwargs[self.session_name] = session
                        result = await run_sync_or_async(handler, *args, **kwargs)
                await self.disconnect() #disconnects from db
                return result
            return wrapper
        return decorator
    
    @property
    def readonly_session(self) -> Callable:
        """
        Returns a decorator that:
        - Creates a new AsyncSession per request.
        - Injects it into the kwargs of the request with the key "session" or custom session name.
        - Closes the session automatically afterward.
        - Does not commit or rollback, for read-only operations.
        """
        def decorator(handler: Callable) -> Callable:
            @wraps(handler)
            async def wrapper(*args, **kwargs):
                if not self._session_factory:
                    raise RuntimeError(
                        "Database is not connected. "
                        "Connection should be established automatically."
                        "Please check network connection and configurations."
                    )
                async with self._session_factory() as session:  # Ensures session closure
                    kwargs[self.session_name] = session
                    return await run_sync_or_async(handler, *args, **kwargs)
            return wrapper
        return decorator



"""
User management CLI
"""
from pyjolt.cli import CLIController, command, argument
from pyjolt.database.sql import AsyncSession
from pyjolt.auth import AuthUtils

from app.extensions import db
from app.api.schemas.user_schemas import UserRegisterSchema
from app.api.models.user import User

class UserManagamentCli(CLIController):
    """
    Methods for user management
    """
    @command("add-user", help="Add a new user")
    @argument("email", arg_type=str, description="User email")
    @argument("fullname", arg_type=str, description="User fullname")
    @argument("password", arg_type=str, description="User password")
    @argument("confirm_password", arg_type=str, description="Confirm user password")
    @db.managed_session_for_cli
    #pylint: disable-next=R0913,R0917
    async def add_user(self, email: str, fullname: str,
                       password: str, confirm_password: str,
                       session: AsyncSession):
        """
        Adds a new user
        """
        user_data = UserRegisterSchema.model_validate({
            "email": email,
            "fullname": fullname,
            "password": password,
            "confirm_password": confirm_password
        }).model_dump()
        if user_data["password"] != user_data["confirm_password"]:
            #pylint: disable-next=W0719
            raise Exception("Password and confirm password must match")
        pass_hash: str = AuthUtils.create_password_hash(password)
        user: User = User(email=email,
                          fullname=fullname,
                          password=pass_hash) # type: ignore
        session.add(user)

        print("User added successfully")



"""Overview extras for different db dialects"""
from typing import Any, Dict, List, Tuple
from sqlalchemy import text
from sqlalchemy.inspection import inspect

def _portable_schema_stats(sync_conn, schema: str | None) -> Dict[str, Any]:
    insp = inspect(sync_conn)

    # pick schemas
    schemas = [schema] if schema else insp.get_schema_names()

    total_tables = 0
    total_views = 0
    total_indexes = 0
    total_foreign_keys = 0
    total_columns = 0
    tables_seen: List[Tuple[str, str]] = []  # (schema, table)

    # collect per-table info
    for sch in schemas:
        tables = insp.get_table_names(schema=sch)
        views = insp.get_view_names(schema=sch)
        total_tables += len(tables)
        total_views += len(views)

        for t in tables:
            tables_seen.append((sch, t))
            # columns
            cols = insp.get_columns(t, schema=sch)
            total_columns += len(cols)
            # indexes
            idxs = insp.get_indexes(t, schema=sch)
            total_indexes += len(idxs)
            # fks
            fks = insp.get_foreign_keys(t, schema=sch)
            total_foreign_keys += len(fks)

    avg_cols_per_table = (total_columns / total_tables) if total_tables else 0.0

    # Detect unindexed FKs (heuristic, portable)
    # For each FK on child table, check if any index covers exactly the fk column set (order-insensitive).
    unindexed_fk_count = 0
    for sch, t in tables_seen:
        fks = insp.get_foreign_keys(t, schema=sch)
        if not fks:
            continue
        idxs = insp.get_indexes(t, schema=sch)
        indexed_sets = {tuple(sorted(ix["column_names"])) for ix in idxs if ix.get("column_names")}
        for fk in fks:
            cols = fk.get("constrained_columns") or []
            if not cols:
                continue
            if tuple(sorted(cols)) not in indexed_sets:
                unindexed_fk_count += 1

    return {
        "schemas_count": len(schemas),
        "tables_count": total_tables,
        "views_count": total_views,
        "indexes_count": total_indexes,
        "foreign_keys_count": total_foreign_keys,
        "columns_count": total_columns,
        "avg_columns_per_table": avg_cols_per_table,
        "unindexed_foreign_keys": unindexed_fk_count,
    }


def _extras_postgres(sync_conn) -> Dict[str, Any]:
    # db size
    size = sync_conn.execute(text("SELECT pg_database_size(current_database())")).scalar()
    # fast row estimates per table
    rows = sync_conn.execute(text("""
        SELECT n.nspname AS schema, c.relname AS table, c.reltuples::bigint AS est_rows
        FROM pg_class c
        JOIN pg_namespace n ON n.oid = c.relnamespace
        WHERE c.relkind = 'r'  -- ordinary tables
          AND n.nspname NOT IN ('pg_catalog','information_schema','pg_toast')
          AND n.nspname NOT LIKE 'pg_%'
    """)).mappings().all()
    per_table_est = {f"{r['schema']}.{r['table']}": int(r["est_rows"]) for r in rows}
    return {
        "db_size_bytes": int(size or 0),
        "row_estimates": per_table_est,
        "row_estimates_total": sum(per_table_est.values()),
        "notes": "Row counts are estimates from pg_class.reltuples.",
    }

def _extras_sqlite(sync_conn) -> Dict[str, Any]:
    # PRAGMAs return page size and page count; multiply for file size bytes
    page_size = sync_conn.execute(text("PRAGMA page_size;")).scalar()
    page_count = sync_conn.execute(text("PRAGMA page_count;")).scalar()
    freelist = sync_conn.execute(text("PRAGMA freelist_count;")).scalar()
    size_bytes = int((page_size or 0) * (page_count or 0))
    return {
        "db_size_bytes": size_bytes,
        "sqlite_page_size": page_size,
        "sqlite_page_count": page_count,
        "sqlite_freelist_pages": freelist,
    }

def _extras_mysql(sync_conn) -> Dict[str, Any]:
    # Use current database()
    db = sync_conn.execute(text("SELECT DATABASE()")).scalar()
    if not db:
        return {}
    rows = sync_conn.execute(text("""
        SELECT
          TABLE_SCHEMA, TABLE_NAME,
          COALESCE(TABLE_ROWS,0) AS table_rows,
          COALESCE(DATA_LENGTH,0) + COALESCE(INDEX_LENGTH,0) AS total_bytes
        FROM information_schema.TABLES
        WHERE TABLE_SCHEMA = :db AND TABLE_TYPE='BASE TABLE'
    """), {"db": db}).mappings().all()
    size_total = sum(int(r["total_bytes"]) for r in rows)
    row_est_total = sum(int(r["table_rows"]) for r in rows)
    per_table_est = {f"{r['TABLE_SCHEMA']}.{r['TABLE_NAME']}": int(r["table_rows"]) for r in rows}
    return {
        "db_size_bytes": size_total,
        "row_estimates": per_table_est,
        "row_estimates_total": row_est_total,
        "notes": "Row counts are estimates from information_schema.TABLES.TABLE_ROWS.",
    }

def _extras_mssql(sync_conn) -> Dict[str, Any]:
    # total database size (approx) and row counts
    # size in 8KB pages
    db = sync_conn.execute(text("SELECT DB_NAME()")).scalar()
    if not db:
        return {}
    size_pages = sync_conn.execute(text("""
        SELECT SUM(size) FROM sys.master_files WHERE database_id = DB_ID()
    """)).scalar()
    size_bytes = int((size_pages or 0) * 8 * 1024)

    # row counts via sys.dm_db_partition_stats (user tables only)
    rows = sync_conn.execute(text("""
        SELECT s.name AS schema_name, t.name AS table_name,
               SUM(p.row_count) AS row_count
        FROM sys.tables t
        JOIN sys.schemas s ON t.schema_id = s.schema_id
        JOIN sys.dm_db_partition_stats p ON t.object_id = p.object_id
        WHERE p.index_id IN (0,1) -- heap or clustered index
        GROUP BY s.name, t.name
    """)).mappings().all()
    per = {f"{r['schema_name']}.{r['table_name']}": int(r["row_count"]) for r in rows}
    return {
        "db_size_bytes": size_bytes,
        "row_counts": per,
        "row_counts_total": sum(per.values()),
        "notes": "Row counts from sys.dm_db_partition_stats; size from sys.master_files.",
    }

def _extras_oracle(sync_conn) -> Dict[str, Any]:
    # NUM_ROWS in ALL_TABLES are stats-based estimates (ANALYZE needed).
    rows = sync_conn.execute(text("""
        SELECT owner AS schema, table_name, COALESCE(num_rows,0) AS est_rows
        FROM ALL_TABLES
    """)).mappings().all()
    per = {f"{r['SCHEMA']}.{r['TABLE_NAME']}": int(r["EST_ROWS"]) for r in rows}
    # segment sizes
    seg = sync_conn.execute(text("""
        SELECT owner AS schema, segment_name, SUM(bytes) AS bytes
        FROM ALL_SEGMENTS
        GROUP BY owner, segment_name
    """)).mappings().all()
    size_total = sum(int(r["BYTES"]) for r in seg)
    return {
        "db_size_bytes": size_total,
        "row_estimates": per,
        "row_estimates_total": sum(per.values()),
        "notes": "Row counts are estimates from ALL_TABLES.NUM_ROWS; size from ALL_SEGMENTS.",
    }



"""Authentication middleware"""
from typing import Any, cast
from pyjolt.auth import Authentication, AuthUtils
from pyjolt import Request
from pyjolt.database.sql import AsyncSession
from app.extensions import db
from app.api.models.user import User

class AuthMW(Authentication):
    """Middleware implementation"""

    configs_name: str = "AUTH_MW"

    async def user_loader(self, req: Request) -> Any:
        """Loads user from the provided cookie"""
        secret_key: str = cast(str, self.app.get_conf("SECRET_KEY"))
        cookie_header = req.headers.get("cookie", "")
        if cookie_header:
            # Split the cookie string on semicolons and equals signs to extract individual cookies
            cookies = dict(cookie.strip().split('=', 1) for cookie in cookie_header.split(';'))
            auth_cookie = cookies.get(self.app.get_conf("COOKIE_NAME"))
            if auth_cookie:
                user_id = AuthUtils.decode_signed_cookie(auth_cookie, secret_key)
                if user_id:
                    return await self.get_user(user_id, session=cast(AsyncSession,None))
        return None

    @db.readonly_session
    async def get_user(self, user_id, session: AsyncSession) -> Any:
        """Loads the user"""
        return await User.query(session).filter_by(id=user_id).first()

    async def role_check(self, user: Any, roles: list[Any]) -> bool:
        return False



"""
Helpers and constants for admin dashboard
"""
from typing import Optional, Type
from enum import StrEnum
from sqlalchemy.inspection import inspect
from ..database.sql.declarative_base import DeclarativeBaseModel

class FormType(StrEnum):
    CREATE = "create"
    UPDATE = "update"

class PermissionType(StrEnum):
    CAN_ENTER = "enter"
    CAN_VIEW = "view"
    CAN_CREATE = "create"
    CAN_UPDATE = "update"
    CAN_DELETE = "delete"

def extract_table_columns(Model: Type[DeclarativeBaseModel], exclude: Optional[list[str]] = None, limit: int = 8):
    """
    Extracts SQLAlchemy model columns suitable for displaying in a generic table.

    Rules:
      - Always include the primary key column(s) first.
      - Then include up to (limit - number_of_pk) additional columns
        sorted alphabetically by their attribute name.
      - Primary key(s) always appear first, even if alphabetically later.

    Args:
        Model: SQLAlchemy declarative model class.
        limit (int): Maximum number of columns to include (default = 8).

    Returns:
        list[Column]: Ordered list of SQLAlchemy Column objects.
    """
    if exclude is None:
        exclude = []
    mapper = inspect(Model)
    cols = list(mapper.columns)

    pk_keys = [c.key for c in cols if c.primary_key]
    non_pk_keys = sorted([c.key for c in cols if not c.primary_key and c.key.lower() not in exclude], key=str.lower)
    take = max(0, limit - len(pk_keys))
    return pk_keys + non_pk_keys[:take]

def register_model(Model: Type[DeclarativeBaseModel]):
    """
    Registers the model with the admin dashboard. Only
    registered models will appear in the dashboard.
    """
    setattr(Model, "__use_in_dashboard__", True)
    return Model



"""
User response models
"""
from typing import Optional
from pydantic import BaseModel, EmailStr, field_validator, Field
from .base_schema import BaseResponse

class LoginSchema(BaseModel):
    """Login schema"""
    email: EmailStr
    password: str
    remember: Optional[bool] = Field(False, description="Long duration cookie")

    @field_validator("password")
    @classmethod
    def check_min_length(cls, value):
        """Min length check"""
        min_length = 8  # for example
        if len(value) < min_length:
            raise ValueError(f"Email must be at least {min_length} characters long.")
        return value

class UserRegisterSchema(BaseModel):
    """User registration schema. Used in CLI"""
    email: EmailStr
    fullname: str
    password: str
    confirm_password: str

    @field_validator("password")
    @classmethod
    def check_min_length(cls, value):
        """Min length check"""
        min_length = 8  # for example
        if len(value) < min_length:
            raise ValueError(f"Email must be at least {min_length} characters long.")
        return value

class UserOutSchema(BaseModel):
    """Standard user out model"""
    id: int
    fullname: str
    email: str

class LoginOutSchema(BaseResponse):
    """Schema for successful login"""
    data: UserOutSchema



"""
Test api endpoint
"""
from typing import Any
from pyjolt import Request, Response, HttpStatus, MediaType, abort
from pyjolt.controller import Controller, path, get, post, consumes, produces
from pydantic import BaseModel, field_serializer
from pyjolt.database.sql import AsyncSession

from app.extensions import db
from app.api.models import Example

class CreateExample(BaseModel):
    name: str

class ExampleOut(CreateExample):
    id: int

    @classmethod
    def from_model(cls, model: Example):
        return cls(id=model.id, name=model.name)

class ResponseModel(BaseModel):
    message: str
    data: Any|None = None

    @field_serializer("data")
    def serialize_data(self, data: Any, _info):
        if isinstance(data, BaseModel):
            return data.model_dump()
        return data

@path("/api/v1/examples")
class ExampleApi(Controller):

    @get("/<int:example_id>")
    @produces(MediaType.APPLICATION_JSON)
    @db.managed_session
    async def get_example(self, req: Request, example_id: int, session: AsyncSession) -> Response[ResponseModel]:
        #The request object is always injected into the endpoint as the first argument
        #All other data (route parameters, indicated data objects etc) are injected
        #in the order of top->bottom, left->right

        example: Example = await Example.query(session).filter_by(id=example_id).first()
        if example is None:
            abort(f"Test with id {example_id} does not exist.", HttpStatus.NOT_FOUND, "error")

        return req.response.json({
            "message": "Test fetched successfully",
            "data": ExampleOut.from_model(example)
        }).status(HttpStatus.OK)
    
    @post("/")
    @consumes(MediaType.APPLICATION_JSON)
    @produces(MediaType.APPLICATION_JSON)
    @db.managed_session
    async def create_example(self, req: Request, data: CreateExample, session: AsyncSession) -> Response[CreateExample]:
        #CreateTest data is validated and injected into the endpoint controlelr
        #Some logic for creating/storing the test object data
        #Response data is serialized as the indicated CreateTest pydantic model in the response type
        #-> Response[CreateTest] indicates what type the response body is
        example: Example = Example(name=data.name)
        session.add(example)
        return req.response.json({
            "name": data.name
        }).status(HttpStatus.CREATED)



"""Admin dashboard package."""
from .admin_dashboard import AdminDashboard, AdminConfig
from .common_controller import AdminEnterError
from .database_controller import AdminPermissionError, UnknownModelError
from .utilities import register_model, PermissionType

__all__ = ["AdminDashboard", "AdminConfig", "PermissionType",
           "AdminPermissionError", "UnknownModelError",
           "AdminEnterError", "register_model"]



"""
Protocls for async NoSQL backends.
"""
from abc import ABC, abstractmethod
from typing import Optional, Callable, Any, Iterable, Mapping, TYPE_CHECKING

if TYPE_CHECKING:
    from ....pyjolt import PyJolt

class AsyncNoSqlBackendBase(ABC):
    """
    Minimal async adapter interface a backend must implement.
    """

    @classmethod
    @abstractmethod
    def configure_from_app(cls, app: "PyJolt", configs: dict[str, Any]) -> "AsyncNoSqlBackendBase":
        """
        Classmethod to configure backend from app config.
        Called during NoSqlDatabase.init_app().
        """
        ...

    @abstractmethod
    async def connect(self) -> None:
        ...

    @abstractmethod
    async def disconnect(self) -> None:
        ...

    @abstractmethod
    def database_handle(self) -> Any:
        """
        Returns an object representing the 'database' to use inside handlers.
        For backends without a database concept, return a client/root handle.
        """
        ...

    @abstractmethod
    def supports_transactions(self) -> bool:
        ...

    @abstractmethod
    async def start_session(self) -> Any:
        """
        Return a session/context object usable in transactions (or None if unsupported).
        """
        ...

    @abstractmethod
    async def with_transaction(self, fn: Callable[..., Any], *args, session: Any = None, **kwargs) -> Any:
        """
        Execute fn inside a transaction if supported; otherwise call fn directly.
        """
        ...

    @abstractmethod
    def get_collection(self, name: str) -> Any:
        ...

    @abstractmethod
    async def find_one(self, collection: str, filter: Mapping[str, Any], *, session: Any = None, **kwargs) -> Any:
        ...

    @abstractmethod
    async def find_many(self, collection: str, filter: Mapping[str, Any] | None = None, *, session: Any = None,
                        limit: Optional[int] = None, skip: Optional[int] = None, sort: Optional[Iterable[tuple[str, int]]] = None,
                        **kwargs) -> list[Any]:
        ...

    @abstractmethod
    async def insert_one(self, collection: str, doc: Mapping[str, Any], *, session: Any = None, **kwargs) -> Any:
        ...

    @abstractmethod
    async def insert_many(self, collection: str, docs: Iterable[Mapping[str, Any]], *, session: Any = None, **kwargs) -> Any:
        ...

    @abstractmethod
    async def update_one(self, collection: str, filter: Mapping[str, Any], update: Mapping[str, Any], *,
                         upsert: bool = False, session: Any = None, **kwargs) -> Any:
        ...

    @abstractmethod
    async def delete_one(self, collection: str, filter: Mapping[str, Any], *, session: Any = None, **kwargs) -> Any:
        ...

    @abstractmethod
    async def aggregate(self, collection: str, pipeline: Iterable[Mapping[str, Any]], *,
                        session: Any = None, **kwargs) -> list[Any]:
        ...

    @abstractmethod
    async def execute_raw(self, *args, **kwargs) -> Any:
        """
        Backend escape hatch for commands that don't fit the generic surface.
        For MongoDB, this could be db.command(...), collection.bulk_write(...), etc.
        """
        ...


"""
Simple run script
"""

if __name__ == "__main__":
    import uvicorn
    uvicorn.run(
        app="app:App",
        host="localhost",
        port=8080,
        lifespan="on",
        reload=True,
        factory=True
    )



"""Common controller methods"""
# pylint: disable=W0719,W0212
from __future__ import annotations

from typing import TYPE_CHECKING, Any
from ..exceptions.http_exceptions import BaseHttpException
from ..http_statuses import HttpStatus
from ..request import Request
from ..response import Response
from ..controller import Controller

if TYPE_CHECKING:
    from .admin_dashboard import AdminDashboard

class AdminEnterError(BaseHttpException):
    """Error for when a user does not have permission to enter the dashboard"""
    def __init__(self, user):
        super().__init__("User doesn't have access to admin dashboard",
                         HttpStatus.UNAUTHORIZED, "error", user)

class CommonAdminController(Controller):
    """Admin dashboard controller."""

    _dashboard: "AdminDashboard"

    def get_common_variables(self) -> dict[str, Any]:
        """Creates dictionary with common templating variables"""
        variables: dict[str, Any] = {
            "dashboard": self.dashboard,
            "configs": self.dashboard.configs,
            "all_dbs": self.dashboard.all_dbs,
            "database_models": self.dashboard._databases_models,
            "email_clients": self.dashboard.email_clients,
            "task_managers": self.dashboard.task_managers
        }
        return variables

    async def cant_enter_response(self, req: Request) -> Response:
        """Response for when a user cannot enter the dashboard"""
        return (await req.res.html(
            "/__admin_templates/denied_entry.html"
        )).status(HttpStatus.UNAUTHORIZED)

    async def can_enter(self, req: Request) -> bool:
        """
        Method for checking permission to enter
        admin dashboard
        """
        has_permission: bool = False
        has_permission = await self.dashboard.has_enter_permission(req)
        if not has_permission:
            raise AdminEnterError(req.user)
        return has_permission

    async def extension_not_available(self, req: Request, extension_type: str):
        """Raises an error for unavailable extension"""
        return await req.res.html(
            "/__admin_template/unavailable_extension.html", {
                "extension_name": extension_type
            }
        )

    @property
    def dashboard(self) -> "AdminDashboard":
        return self._dashboard



"""
Schedulers related admin dashboard controller
"""
from ..task_manager import TaskManager
from ..request import Request
from ..response import Response
from ..http_statuses import HttpStatus
from ..controller import get
from .common_controller import CommonAdminController
from ..utilities import run_in_background

class AdminTaskManagersController(CommonAdminController):

    @get("/task-managers")
    async def task_managers(self, req: Request) -> Response:
        """List of all schedulers"""
        return await req.res.html(
            "/__admin_templates/task_managers.html", {
                "task_managers": self.dashboard.get_task_managers(),
                **self.get_common_variables()
            }
        )
    
    @get("/task-managers/<string:manager_name>")
    async def task_manager(self, req: Request, manager_name: str) -> Response:
        """Selected task manager"""
        managers: dict[str, TaskManager]|None = self.dashboard.get_task_managers()
        if managers is None:
            return await self.extension_not_available(req, "TaskManager")

        manager: list[TaskManager] = list(filter(lambda mng: mng.configs_name == manager_name, managers.values()))
        if len(manager) == 0:
            return await self.extension_not_available(req, "TaskManager - " + manager_name)
        return await req.res.html(
            "/__admin_templates/task_manager.html", {
                "manager": manager[0],
                "manager_name": manager[0].configs_name,
                "tasks": manager[0].jobs,
                **self.get_common_variables()
            }
        )

    #API calls for task management
    @get("/task-managers/<string:manager_name>/run/<string:task_id>")
    async def run_task(self, req: Request, manager_name: str, task_id: str) -> Response:
        """Runs task with provided ID"""
        try:
            managers: dict[str, TaskManager]|None = self.dashboard.get_task_managers()
            if managers is None:
                return await self.extension_not_available(req, "TaskManager")
            manager: list[TaskManager] = list(filter(lambda mng: mng.configs_name == manager_name, managers.values()))
            if len(manager) == 0:
                return await self.extension_not_available(req, "TaskManager - " + manager_name)
            
            task = manager[0].jobs.get(task_id)
            if task is None:
                return req.res.json({
                    "message": f"Task with id '{task_id}' does not exist",
                    "status": "danger"
                }).status(HttpStatus.BAD_REQUEST)
            self.app.logger.info(f"Running task with {task_id=} in task manager {manager_name}")
            run_in_background(task.func)
            return req.res.json({
                "message": "Task started",
                "status": "success"
            }).status(HttpStatus.OK)
        except Exception:
            return req.res.json({
                "message": f"Something went wrong. Make sure task manager {manager_name} and task {task_id=} exist.",
                "status": "danger"
            }).status(HttpStatus.INTERNAL_SERVER_ERROR)

    @get("/task-managers/<string:manager_name>/pause/<string:task_id>")
    async def pause_task(self, req: Request, manager_name: str, task_id: str) -> Response:
        """Pauses task with provided ID"""
        try:
            managers: dict[str, TaskManager]|None = self.dashboard.get_task_managers()
            if managers is None:
                return await self.extension_not_available(req, "TaskManager")
            manager: list[TaskManager] = list(filter(lambda mng: mng.configs_name == manager_name, managers.values()))
            if len(manager) == 0:
                return await self.extension_not_available(req, "TaskManager - " + manager_name)

            self.app.logger.info(f"Pausing task with {task_id=} in task manager {manager_name}")
            manager[0].pause_job(task_id)

            return req.res.json({
                "message": "Task paused",
                "status": "success"
            }).status(HttpStatus.OK)
        except Exception:
            return req.res.json({
                "message": f"Something went wrong. Check if job with {task_id=} exists in task manager {manager_name}",
                "status": "danger"
            }).status(HttpStatus.INTERNAL_SERVER_ERROR)
    
    @get("/task-managers/<string:manager_name>/resume/<string:task_id>")
    async def resume_task(self, req: Request, manager_name: str, task_id: str) -> Response:
        """Resumes task with provided ID"""
        try:
            managers: dict[str, TaskManager]|None = self.dashboard.get_task_managers()
            if managers is None:
                return await self.extension_not_available(req, "TaskManager")
            manager: list[TaskManager] = list(filter(lambda mng: mng.configs_name == manager_name, managers.values()))
            if len(manager) == 0:
                return await self.extension_not_available(req, "TaskManager - " + manager_name)

            self.app.logger.info(f"Resuming task with {task_id=} in task manager {manager_name}")
            manager[0].resume_job(task_id)

            return req.res.json({
                "message": "Task started",
                "status": "success"
            }).status(HttpStatus.OK)
        except Exception:
            return req.res.json({
                "message": f"Something went wrong. Check if job with {task_id=} exists in task manager {manager_name}",
                "status": "danger"
            }).status(HttpStatus.INTERNAL_SERVER_ERROR)



"""
Authentication API
"""
from typing import cast
from pyjolt import Response, Request, HttpStatus, MediaType, abort
from pyjolt.controller import (Controller, path, get, post,
                               consumes, produces, open_api_docs,
                               Descriptor)
from pyjolt.database.sql import AsyncSession
from pyjolt.auth import login_required, AuthUtils

from app.api.models.user import User
from app.api.schemas.base_schema import ErrorResponseSchema, BaseResponse
from app.api.schemas.user_schemas import UserOutSchema, LoginSchema, LoginOutSchema
from app.extensions import db

@path("/api/v1/auth")
class AuthApi(Controller):
    """
    Authentication controller
    """

    @get("/current")
    @produces(MediaType.APPLICATION_JSON)
    @open_api_docs(Descriptor(status=HttpStatus.FORBIDDEN,
                              description="Not logged in",
                              body=ErrorResponseSchema))
    @login_required
    async def current_user(self, req: Request) -> Response[LoginOutSchema]:
        """Returns current authenticated user"""
        return req.res.json({
            "message": "User is authenticated",
            "status": "success",
            "data": UserOutSchema(
            id=req.user.id,
            fullname=req.user.fullname,
            email=req.user.email
        )}).status(HttpStatus.OK)

    @post("/admin", open_api_spec=False)
    @consumes(MediaType.APPLICATION_X_WWW_FORM_URLENCODED)
    @produces(MediaType.APPLICATION_JSON)
    async def login_form(self, req: Request,
                         login_data: LoginSchema) -> Response[LoginOutSchema]:
        """Login for admin dashboard"""
        print("Login data: ", login_data.model_dump())
        return await self.login_user(req, login_data)

    @post("/login")
    @consumes(MediaType.APPLICATION_JSON)
    @produces(MediaType.APPLICATION_JSON)
    @open_api_docs(Descriptor(status=HttpStatus.BAD_REQUEST,
                              description="Wrong credentials",
                              body=ErrorResponseSchema))
    async def login(self, req: Request,
                    login_data: LoginSchema) -> Response[LoginOutSchema]:
        """Login route"""
        return await self.login_user(req, login_data)

    @get("/logout")
    @produces(MediaType.APPLICATION_JSON)
    @open_api_docs(Descriptor(status=HttpStatus.FORBIDDEN,
                              description="Must be logged in",
                              body=ErrorResponseSchema))
    @login_required
    async def logout(self, req: Request) -> Response[BaseResponse]:
        """User logout"""
        req.res.set_cookie(cookie_name=self.app.get_conf("COOKIE_NAME"),
                           value="",
                           max_age=0,
                           http_only=True)
        return req.res.json({
            "message": "Logout successful",
            "status": "success"
        }).status(HttpStatus.OK)

    @db.readonly_session
    async def login_user(self, req: Request,
                         login_data: LoginSchema,
                         session: AsyncSession = cast(AsyncSession, None)):
        """User login"""
        user: User = await User.query(session).filter_by(email=login_data.email).first()
        if user is None:
            abort(msg="Wrong credentials.", status_code=HttpStatus.BAD_REQUEST)

        valid_pass: bool = AuthUtils.check_password_hash(login_data.password, user.password)
        if valid_pass is False:
            abort(msg="Wrong credentials", status_code=HttpStatus.BAD_REQUEST)
        secret_key: str = self.app.get_conf("SECRET_KEY")
        cookie: str = AuthUtils.create_signed_cookie_value(user.id, secret_key)
        req.res.set_cookie(cookie_name=self.app.get_conf("COOKIE_NAME"),
                           value=cookie,
                           max_age=self.app.get_conf("COOKIE_DURATION"),
                           http_only=True)
        self.app.logger.info(f"User: {user.fullname}(id={user.id}) logged in.")

        return req.res.json({
            "message": "Login successful",
            "status": "success",
            "data": UserOutSchema(id=user.id,fullname=user.fullname,email=user.email)
        }).status(HttpStatus.OK)



"""
Application extensions
"""
from pyjolt.database.sql import SqlDatabase
from pyjolt.database.sql.migrate import Migrate
from pyjolt.email import EmailClient
from pyjolt.caching import Cache

db: SqlDatabase = SqlDatabase()
migrate: Migrate = Migrate(db)
email: EmailClient = EmailClient()
second_email: EmailClient = EmailClient(configs_name="SECOND_EMAIL")
cache: Cache = Cache()

other_db: SqlDatabase = SqlDatabase(db_name="other_db", configs_name="OTHER_DB")
other_migrate: Migrate = Migrate(other_db)



"""
App configs
"""
import os
from typing import cast

from pyjolt import BaseConfig, LogLevel
from pyjolt.admin import AdminConfig
from pyjolt.database.sql import SqlDatabaseConfig
from pyjolt.email import EmailConfig
from pyjolt.logging import LoggerConfig
from pyjolt.state_machine import StateMachineConfig

from app.api.schemas.state_machine_schema import StateMachineTransitionRequestData

class Config(BaseConfig):
    """
    All application configurations
    """

    APP_NAME: str = cast(str, os.environ.get("APP_NAME", None))
    BASE_PATH: str = os.path.dirname(__file__)
    DEBUG: bool = BaseConfig.value_to_bool(os.environ.get("DEBUG", True))
    SECRET_KEY: str = cast(str, os.environ.get("SECRET_KEY", None))
    VERSION: str = cast(str, os.environ.get("VERSION", None))

    COOKIE_NAME: str = cast(str, os.environ.get("COOKIE_NAME", None))
    COOKIE_DURATION: int = int(cast(int, os.environ.get("COOKIE_DURATION", None)))

    CROSSREF_API_URL: str = cast(str, os.environ.get("CROSSREF_API_URL"))
    INSTITUTE_EMAIL: str = cast(str, os.environ.get("INSTITUTE_EMAIL"))

    CORS_ALLOW_ORIGINS: list[str] = [
        "http://localhost:8080",
    ]

    ADMIN_DASHBOARD: AdminConfig = {
        "URL_FOR_FOR_LOGIN": "AuthApi.login_form",
        "URL_FOR_FOR_LOGOUT": "AuthApi.logout",
    }

    SQL_DATABASE: SqlDatabaseConfig = {
        "DATABASE_URI": cast(str, os.environ.get("DATABASE_URI", None)),
        "ALEMBIC_DATABASE_URI_SYNC": cast(str, os.environ.get("ALEMBIC_DATABASE_URI_SYNC", None)),
        "SHOW_SQL": False,
        "NICE_NAME": "Super cool sqlite db"
    }

    OTHER_DB: SqlDatabaseConfig = {
        "DATABASE_URI": cast(str, os.environ.get("OTHER_DATABASE_URI", None)),
        "ALEMBIC_DATABASE_URI_SYNC": cast(str, os.environ.get("ALEMBIC_OTHER_DATABASE_URI_SYNC", None)),
        "ALEMBIC_MIGRATION_DIR": "other_migrations",
        "SHOW_SQL": False,
        "NICE_NAME": "Other cool sqlite db"
    }

    EMAIL_CLIENT: EmailConfig = {
        "SENDER_NAME_OR_ADDRESS": "info@physio-mb.si",
        "SMTP_SERVER": "localhost",
        "SMTP_PORT": 1025,
        "USE_TLS": False
    }

    SECOND_EMAIL: EmailConfig = {
        "SENDER_NAME_OR_ADDRESS": "newsletter@physio-mb.si",
        "SMTP_SERVER": "localhost",
        "SMTP_PORT": 1025,
        "USE_TLS": False
    }

    STATE_MACHINE: StateMachineConfig = {
        "INCLUDE_OPEN_API": False,
        "USE_AUTH": False,
        "TRANSITION_DATA": StateMachineTransitionRequestData,
        "API_URL": "/api/v1/state-machine"
    }

    MODELS: list[str] = [
        "app.api.models.user:User",
        "app.api.models.post:Post",
        "app.api.models.publication:Publication",
        "app.api.models.project:Project",
        "app.api.models.researcher:Researcher",
        "app.api.models.other_db.dummy:Dummy"
    ]

    CONTROLLERS: list[str] = [
        "app.api.auth.auth_api:AuthApi",
        "app.api.posts.posts_api:PostsApi",
        "app.api.publications.publications_api:PublicationsApi",
        "app.api.projects.projects_api:ProjectsApi",
        "app.api.researchers.researchers_api:ResearchersApi"
    ]

    CLI_CONTROLLERS: list[str] = [
        "app.cli.user_management:UserManagamentCli",
        "app.cli.dev_utils:DevUtilities"
    ]

    EXCEPTION_HANDLERS: list[str] = [
        "app.api.exception_handlers.exception_handler:CustomExceptionHandler"
    ]

    EXTENSIONS: list[str] = [
        "app.extensions:db",
        "app.extensions:migrate",
        "app.extensions:other_db",
        "app.extensions:other_migrate",
        "app.extensions:email",
        "app.extensions:second_email",
        "app.task_manager:scheduler_manager",
        "app.state_machine:state_machine",
        "app.admin:admin_extension"
    ]

    # LOGGERS: list[str] = [
    #     "app.loggers.performance_logger:PerformanceFileLogger"
    # ]

    MIDDLEWARE: list[str] = [
        "app.middleware.timing_mw:TimingMW",
        "app.middleware.auth_mw:AuthMW"
    ]

    PERFORMANCE_FILE_LOGGER: LoggerConfig = {
        "SINK": os.path.join(BASE_PATH, "logging", "performance_log.log"),
        "LEVEL": LogLevel.INFO,
        "ENQUEUE": True,
        "BACKTRACE": True,
        "DIAGNOSE": True,
        "COLORIZE": False,
        "DELAY": True,
        "ROTATION": "5 MB",
        "RETENTION": 5,
        "COMPRESSION": "zip",
        "SERIALIZE": False,
        "ENCODING": "utf-8",
        "MODE": "a",
    }



"""HTTP methods enum"""
from enum import StrEnum

class HttpMethod(StrEnum):
    """All allowed http methods"""
    GET = "GET"
    POST = "POST"
    PUT = "PUT"
    PATCH = "PATCH"
    DELETE = "DELETE"
    SOCKET = "SOCKET"



"""
Default static endpoint that serves all static files for the application
In production, static files should be serves directly by a reverse proxy server such
as Nginx. This reverse proxy server approach is more efficient
"""
import os
import mimetypes
from typing import TYPE_CHECKING
from werkzeug.security import safe_join

from .exceptions import StaticAssetNotFound
from .controller import Controller, get
from .utilities import get_file, get_range_file

if TYPE_CHECKING:
    from .request import Request

class Static(Controller):

    @get("/<path:filename>")
    async def get(self, req: "Request", filename: str):
        """
        Endpoint for static files with HTTP Range support,
        falling back to get_file for full-content requests.
        """
        # Checks if file exists
        file_path = None
        candidate = safe_join(req.app.static_files_path, filename)
        if candidate and os.path.exists(candidate):
            file_path = candidate
        if not file_path:
            raise StaticAssetNotFound()

        # checks/guesses mimetype
        guessed, _ = mimetypes.guess_type(file_path)
        content_type = guessed or "application/octet-stream"

        # Checks range header and returns range if header is present
        range_header = req.headers.get("range")
        if not range_header:
            status, headers, body = await get_file(file_path, content_type=content_type)
            headers["Accept-Ranges"] = "bytes"
            return req.res.send_file(body, headers).status(status)

        return await get_range_file(req.res, file_path, range_header, content_type)



"""
Timing middleware
"""
import time
from pyjolt import Request, Response
from pyjolt.middleware import MiddlewareBase

class TimingMW(MiddlewareBase):
    """Timing middleware handler"""
    async def middleware(self, req: Request) -> Response:
        t0: float = time.perf_counter()
        res = await self.next(req) # pass down
        self.app.logger.info(
            "PERFORMANCE: Request to "
            f"{req.path}?{req.query_string} "
            f"took {int((time.perf_counter() - t0)*1000)} ms"
        )
        return res



"""
SQLite cache backend

Indicated values are defaults
CACHE_SQLITE_PATH: str = "./pyjolt_cache.sqlite3" # file path (required for persistence)
CACHE_SQLITE_TABLE: str = "cache_entries" # table name
CACHE_KEY_PREFIX: str = "pyjolt:cache:" # optional key namespace
CACHE_DURATION: int = 300 # default TTL seconds
CACHE_SQLITE_WAL_CHECKPOINT_MODE: str = "PASSIVE" #PASSIVE|FULL|RESTART|TRUNCATE
CACHE_SQLITE_WAL_CHECKPOINT_EVERY: int = 100 #run checkpoint every N write ops
"""
from __future__ import annotations

import os
import pickle
import time
from typing import Optional, TYPE_CHECKING, Tuple, cast, Any
from pydantic import BaseModel, Field

import aiosqlite

from .base_cache_backend import BaseCacheBackend

if TYPE_CHECKING:
    from ...pyjolt import PyJolt

class SQLiteCacheConfig(BaseModel):
    CACHE_SQLITE_PATH: str = Field("./pyjolt_cache.db", description="SQLite DB file path (use ':memory:' for in-memory)")
    CACHE_SQLITE_TABLE: str = Field("cache_entries", description="Table name for cache entries")
    CACHE_DURATION: int = Field(300, description="Cache default TTL in seconds")
    CACHE_KEY_PREFIX: str = Field("", description="Cache key prefix/namespace")
    CACHE_SQLITE_WAL_CHECKPOINT_MODE: str = Field("PASSIVE", description="Mode for WAL checkpointing: PASSIVE|FULL|RESTART|TRUNCATE")
    CACHE_SQLITE_WAL_CHECKPOINT_EVERY: int = Field(100, description="Insert WAL checkpoint every N write operations")

class SQLiteCacheBackend(BaseCacheBackend):
    """SQLite-backed cache using pickled payloads (async via aiosqlite)."""

    def __init__(
        self,
        db_path: str,
        table: str = "cache_entries",
        default_ttl: int = 300,
        key_prefix: str = "",
        checkpoint_mode: str = "PASSIVE",
        checkpoint_every: int = 100,
    ) -> None:
        if not db_path:
            raise ValueError("CACHE_SQLITE_PATH must be provided for SQLiteCacheBackend")
        self._db_path = db_path
        self._table = table
        self.default_ttl = int(default_ttl)
        self._conn: Optional[aiosqlite.Connection] = None
        if key_prefix and not key_prefix.endswith(":"):
            key_prefix = key_prefix + ":"
        self._prefix = key_prefix
        # WAL checkpoint settings
        self._checkpoint_mode = checkpoint_mode.upper()
        if self._checkpoint_mode not in {"PASSIVE", "FULL", "RESTART", "TRUNCATE"}:
            raise ValueError("CACHE_SQLITE_WAL_CHECKPOINT_MODE must be one of PASSIVE|FULL|RESTART|TRUNCATE")
        self._checkpoint_every = max(1, int(checkpoint_every))
        self._write_ops = 0

    @classmethod
    def configure_from_app(cls, app: "PyJolt", configs: dict[str, Any]) -> "SQLiteCacheBackend":
        db_path = cast(str, configs.get("SQLITE_PATH", "./pyjolt_cache.db"))
        table = cast(str, configs.get("SQLITE_TABLE", "cache_entries"))
        default_ttl = int(configs.get("DURATION", 300))
        key_prefix = cast(str, configs.get("KEY_PREFIX", ""))
        checkpoint_mode = cast(str, configs.get("SQLITE_WAL_CHECKPOINT_MODE", "PASSIVE"))
        checkpoint_every = int(configs.get("SQLITE_WAL_CHECKPOINT_EVERY", 100))
        if db_path != ":memory:":
            os.makedirs(os.path.dirname(os.path.abspath(db_path)), exist_ok=True)
        return cls(
            db_path=db_path,
            table=table,
            default_ttl=default_ttl,
            key_prefix=key_prefix,
            checkpoint_mode=checkpoint_mode,
            checkpoint_every=checkpoint_every,
        )

    async def connect(self) -> None:
        if self._conn:
            return
        self._conn = await aiosqlite.connect(self._db_path)
        # Pragmas tuned for app caches (good perf/safety tradeoffs)
        await self._conn.execute("PRAGMA journal_mode=WAL;")
        await self._conn.execute("PRAGMA synchronous=NORMAL;")
        await self._conn.execute("PRAGMA foreign_keys=ON;")
        await self._conn.execute("PRAGMA busy_timeout=3000;")# 3s wait instead of immediate 'database is locked'. Useful for WAL mode with concurrent readers/writers (multiple app workers)
        await self._conn.execute("PRAGMA wal_autocheckpoint=1000;")# checkpoint every ~1000 pages (~4MB)
        await self._ensure_schema()
        await self._conn.commit()

    async def disconnect(self) -> None:
        if self._conn is None:
            return
        await self._conn.close()
        self._conn = None

    async def _ensure_schema(self) -> None:
        assert self._conn is not None
        t = self._table
        await self._conn.executescript(
            f"""
            CREATE TABLE IF NOT EXISTS {t} (
                k TEXT PRIMARY KEY,
                v BLOB NOT NULL,
                expire REAL NOT NULL,
                updated_at REAL NOT NULL
            );
            CREATE INDEX IF NOT EXISTS idx_{t}_expire ON {t}(expire);
            CREATE INDEX IF NOT EXISTS idx_{t}_k_pref ON {t}(k);
            """
        )

    def _k(self, key: str) -> str:
        return f"{self._prefix}{key}" if self._prefix else key

    async def _cleanup_expired(self) -> None:
        assert self._conn is not None
        now = time.time()
        await self._conn.execute(f"DELETE FROM {self._table} WHERE expire <= ?", (now,))
        await self._conn.commit()

    async def _maybe_checkpoint(self) -> None:
        """Run WAL checkpoint every N write operations."""
        assert self._conn is not None
        self._write_ops += 1
        if (self._write_ops % self._checkpoint_every) == 0:
            # PRAGMA wal_checkpoint(PASSIVE)
            await self._conn.execute(f"PRAGMA wal_checkpoint({self._checkpoint_mode});")
            await self._conn.commit()

    async def get(self, key: str) -> Optional[dict]:
        if self._conn is None:
            await self.connect()
        assert self._conn is not None
        k = self._k(key)
        async with self._conn.execute(
            f"SELECT v, expire FROM {self._table} WHERE k = ?",
            (k,),
        ) as cur:
            row = await cur.fetchone()
        if row is None:
            return None
        raw, expire = cast(Tuple[bytes, float], row)
        now = time.time()
        if expire <= now:
            await self._conn.execute(f"DELETE FROM {self._table} WHERE k = ?", (k,))
            await self._conn.commit()
            # count deletion as a write op for checkpoint cadence
            await self._maybe_checkpoint()
            return None
        try:
            return cast(dict, pickle.loads(raw))
        except Exception:
            # Drop corrupt entry
            await self._conn.execute(f"DELETE FROM {self._table} WHERE k = ?", (k,))
            await self._conn.commit()
            await self._maybe_checkpoint()
            return None

    async def set(self, key: str, value: dict, duration: Optional[int] = None) -> None:
        if self._conn is None:
            await self.connect()
        assert self._conn is not None
        ttl = int(duration) if duration is not None else self.default_ttl
        expire = time.time() + ttl
        k = self._k(key)
        raw = pickle.dumps(value)
        # Upsert with ON CONFLICT
        await self._conn.execute(
            f"""
            INSERT INTO {self._table}(k, v, expire, updated_at)
            VALUES(?, ?, ?, ?)
            ON CONFLICT(k) DO UPDATE SET
                v=excluded.v,
                expire=excluded.expire,
                updated_at=excluded.updated_at
            """,
            (k, raw, expire, time.time()),
        )
        await self._conn.commit()
        # count insertion as a write op for checkpoint cadence
        await self._cleanup_expired()
        await self._maybe_checkpoint()

    async def delete(self, key: str) -> None:
        if self._conn is None:
            await self.connect()
        assert self._conn is not None
        await self._conn.execute(f"DELETE FROM {self._table} WHERE k = ?", (self._k(key),))
        await self._conn.commit()
        await self._maybe_checkpoint()

    async def clear(self) -> None:
        if self._conn is None:
            await self.connect()
        assert self._conn is not None
        if self._prefix:
            like = f"{self._prefix}%"
            await self._conn.execute(f"DELETE FROM {self._table} WHERE k LIKE ?", (like,))
        else:
            await self._conn.execute(f"DELETE FROM {self._table}")
        #Checkpoint after full clear to truncate WAL
        await self._conn.execute(f"PRAGMA wal_checkpoint({self._checkpoint_mode});")
        await self._conn.commit()
        #reset write op count after full clear
        self._write_ops = 0



"""
Init file for PyJolt package
"""

from .pyjolt import PyJolt, app, app_path, on_shutdown, on_startup
from .base_extension import BaseExtension
from .middleware import MiddlewareBase
from .exceptions import abort, html_abort
from .configuration_base import BaseConfig

from .request import Request, UploadedFile
from .response import Response

from .utilities import run_sync_or_async, run_in_background
from .media_types import MediaType
from .http_methods import HttpMethod
from .http_statuses import HttpStatus
from .logging.logger_config_base import LogLevel

__all__ = ['PyJolt', 'abort', 'Request', 'Response',
           'run_sync_or_async', 'run_in_background',
           'UploadedFile', 'MediaType', 'HttpMethod',
           'HttpStatus', 'html_abort',
           'app', 'app_path', 'on_shutdown',
           'on_startup', 'BaseExtension', 'BaseConfig',
           'LogLevel', 'MiddlewareBase']



"""
Base extension class
"""
from typing import TYPE_CHECKING, Any, Optional, cast
from abc import abstractmethod, ABC
from pydantic import BaseModel, ValidationError

if TYPE_CHECKING:
    from .pyjolt import PyJolt

class BaseExtension(ABC):

    _configs_name: str
    _app: "Optional[PyJolt]"
    _configs: dict[str, Any]

    @abstractmethod
    def init_app(self, app: "PyJolt") -> None:
        ...
    
    def validate_configs(self, configs: dict[str, Any], model: type[BaseModel]) -> dict[str, Any]:
        try:
            return model.model_validate(configs).model_dump()
        except ValidationError as e:
            raise ValueError(f"Invalid configuration for {self._configs_name or self.__class__.__name__}: {e}") from e
    
    @property
    def configs_name(self) -> str:
        """
        Return the config name used in app configurations
        for this extension.
        """
        return self._configs_name

    @property
    def app(self) -> "PyJolt":
        if self._app is None:
            raise RuntimeError("Extension not initialized with a PyJolt app.")
        return cast("PyJolt", self._app)
    
    @property
    def configs(self) -> dict[str, Any]:
        """Returns a dictinary of extension configs"""
        return self._configs
    
    @property
    def nice_name(self) -> str|None:
        """Returns nice name of the extension or None"""
        return self._configs.get("NICE_NAME", None)



# request.py
#pylint: disable=C0116
import re
import json
from io import BytesIO
from urllib.parse import parse_qs
from typing import Callable, Any, Union, TYPE_CHECKING, Mapping, cast
import python_multipart as pm

from .response import Response

if TYPE_CHECKING:
    from .pyjolt import PyJolt

def extract_boundary(content_type: str) -> str:
    """
    Pull the boundary=... out of a Content-Type header.
    """
    match = re.search(r'boundary="?([^";]+)"?', content_type)
    if not match:
        raise ValueError("No boundary found in Content-Type")
    return match.group(1)

class UploadedFile:
    """
    Wrapper around an in-memory/temporary file.
    """
    def __init__(self, filename: str, content: bytes, content_type: str):
        self.filename = filename
        self.content_type = content_type
        self._content = content
        self._stream = BytesIO(content)

    def read(self, size: int = -1) -> bytes:
        return self._stream.read(size)

    def seek(self, pos: int, whence: int = 0) -> int:
        return self._stream.seek(pos, whence)

    def save(self, path: str) -> None:
        with open(path, "wb") as f:
            self.seek(0)
            f.write(self._stream.read())

    @property
    def size(self) -> int:
        cur = self._stream.tell()
        self._stream.seek(0, 2)
        sz = self._stream.tell()
        self._stream.seek(cur)
        return sz

    @property
    def stream(self) -> BytesIO:
        self.seek(0)
        return self._stream

    def get_stream(self) -> BytesIO:
        return BytesIO(self._content)

    def __repr__(self) -> str:
        return (f"<UploadedFile filename={self.filename!r} "
                f"size={self.size} content_type={self.content_type!r}>")

class Request:
    """
    ASGI-style request adapter that lazy-parses JSON, form, and multipart.
    """
    def __init__(
        self,
        scope: dict,
        receive: Callable[..., Any],
        app: "PyJolt",
        route_parameters: dict|Mapping,
        route_handler: Callable
    ):
        self._app = app
        self.scope = scope
        self._receive = receive
        self._send: Callable = None  # type: ignore
        self._body: Union[bytes, None] = None
        self._json: Union[dict, None]  = None
        self._form: Union[dict, None]  = None
        self._files: Union[dict, None]  = None
        self._user: Any = None
        self._route_parameters = route_parameters
        self._route_handler    = route_handler
        self._response: Response[Any] = app.response_class(app, self)
        self.state: Any = {}

    @property
    def route_handler(self) -> Callable:
        return self._route_handler

    @property
    def route_parameters(self) -> dict|Mapping:
        return self._route_parameters

    @route_parameters.setter
    def route_parameters(self, rp: dict|Mapping) -> None:
        self._route_parameters = rp

    @property
    def method(self) -> str:
        if self._send is not None:
            return "SOCKET"
        return self.scope.get("method", "").upper()

    @property
    def path(self) -> str:
        return self.scope.get("path", "/")

    @property
    def query_string(self) -> str:
        return self.scope.get("query_string", b"").decode("utf-8")

    @property
    def headers(self) -> dict[str, str]:
        """
        Decode the raw ASGI headers into a dict of lowercase str→str.
        """
        raw = self.scope.get("headers", [])
        return {
            key.decode("latin1").lower(): val.decode("latin1")
            for key, val in raw
        }

    @property
    def query_params(self) -> dict[str, str]:
        qs = self.scope.get("query_string", b"")
        parsed = parse_qs(qs.decode("utf-8"))
        return cast(dict[str, str], {k: v if len(v) > 1 else v[0] for k, v in parsed.items()})

    @property
    def user(self) -> Any:
        return self._user
    
    @property
    def app(self) -> "PyJolt":
        return self._app

    def set_user(self, user: Any) -> None:
        self._user = user

    def remove_user(self) -> None:
        self._user = None

    async def body(self) -> bytes:
        if self._body is not None:
            return self._body

        parts = []
        while True:
            msg = await self._receive()
            if msg["type"] == "http.request":
                parts.append(msg.get("body", b""))
                if not msg.get("more_body", False):
                    break
        self._body = b"".join(parts)
        return self._body

    async def json(self) -> dict[str, Any]|None:
        if self._json is not None:
            return self._json
        raw = await self.body()
        if not raw:
            return None
        try:
            self._json = json.loads(raw)
        except json.JSONDecodeError:
            self._json = None
        return self._json

    async def form(self) -> dict[str, Any]:
        if self._form is not None:
            return self._form

        ct = self.headers.get("content-type", "")
        if "multipart/form-data" in ct:
            self._form, self._files = await self._parse_multipart(ct)
        elif "application/x-www-form-urlencoded" in ct:
            raw = await self.body()
            parsed = parse_qs(raw.decode("utf-8"))
            self._form = {k: v if len(v) > 1 else v[0] for k, v in parsed.items()}
        else:
            self._form = {}

        return self._form

    async def files(self) -> dict[str, UploadedFile]:
        if self._files is None:
            await self.form()
        return self._files or {}

    async def form_and_files(self) -> dict[str, Any]:
        f = await self.form()
        fs = await self.files()
        return {**f, **fs}
    
    async def send(self, message: dict) -> None:
        if self._send is None:
            raise RuntimeError("Send function is available only on websocket requests")
        return await self._send(message)
    
    def set_send(self, send: Callable) -> None:
        self._send = send
    
    async def receive(self) -> dict:
        if self._send is None:
            raise RuntimeError("Receive function is available only on websocket requests")
        return await self._receive()
    
    async def accept(self) -> None:
        if self._send is None:
            raise RuntimeError("Accept function is available only on websocket requests")
        await self._send({"type": "websocket.accept"})

    async def _parse_multipart(self, content_type: str) -> tuple[dict, dict]:
        """
        Stream the body through python-multipart, collecting fields and files.
        """
        raw = await self.body()
        stream = BytesIO(raw)

        form_data: dict[str, str] = {}
        files:     dict[str, UploadedFile] = {}

        def on_field(field: Any) -> None:
            name = field.field_name
            val  = getattr(field, "value", None)

            if isinstance(name, bytes):
                name = name.decode("latin1")
            if isinstance(val, bytes):
                val = val.decode("utf-8", "replace")

            form_data[name] = cast(str, val)

        def on_file(f: Any) -> None:
            raw_name = getattr(f, "field_name", b"") or b""
            raw_fn   = getattr(f, "file_name", b"") or b""
            name = raw_name.decode("latin1") if isinstance(raw_name, bytes) else raw_name
            fn   = raw_fn.decode("latin1")   if isinstance(raw_fn,   bytes) else raw_fn

            fileobj = f.file_object
            fileobj.seek(0)
            content = fileobj.read()

            part_ct = ""
            hdrs = getattr(f, "headers", {})
            if isinstance(hdrs, dict):
                c = hdrs.get(b"Content-Type") or hdrs.get("Content-Type")
                if isinstance(c, bytes):
                    part_ct = c.decode("latin1")
                elif isinstance(c, str):
                    part_ct = c

            files[name] = UploadedFile(
                filename=fn,
                content=content,
                content_type=part_ct
            )

        header_map = {"Content-Type": content_type}

        pm.parse_form(
            headers=cast(dict[str, bytes], header_map),
            input_stream=stream,
            on_field=on_field,
            on_file=on_file,
        )

        return form_data, files

    async def get_data(self, location: str = "json") -> dict[str, Any]|None:
        if location == "json":
            return await self.json()
        if location == "form":
            return await self.form()
        if location == "files":
            return await self.files()
        if location == "form_and_files":
            return await self.form_and_files()
        if location == "query":
            return self.query_params
        return None
    
    @property
    def response(self) -> Response:
        return self._response

    @property
    def res(self) -> Response:
        return self._response



"""
WTForm inputs
"""
from __future__ import annotations
from enum import StrEnum
from typing import cast
from abc import ABC, abstractmethod
from markupsafe import Markup
from wtforms.widgets import html_params

from ..utilities import to_kebab_case

class FormFieldTypes(StrEnum):
    TAGS = "TagsField"
    PASSWORD = "PasswordField"
    EMAIL = "EmailField"
    TEXT = "TextField"
    TEXTAREA = "TextAreaField"
    SELECT = "SelectField"
    RADIO = "RedioField"
    CHECKBOX = "CheckboxField"

class FormField(ABC):

    type: FormFieldTypes = cast(FormFieldTypes, None)

    def __init__(self, **kwargs):
        """
        Initilizer for HTML input elements. Accepts any number of keyword arguments. Each argument will be treated as 
        a html attribute.
        """
        if self.type is None:
            raise NotImplementedError("FormField type must be defined in subclass with a defined type attribute of type FormFieldTypes.")
        self.id = kwargs.pop("id", None)
        self.name = kwargs.pop("name", None)
        self._label = kwargs.pop("label", None)
        self.kwargs = kwargs

        if self.id is None or self.name is None:
            raise Exception("FormField element must have a valid id and name parameter.")

    def __call__(self, **kwargs):
        """
        Render: <tags-input ...standard attrs... as-string="..."></tags-input>
        """
        kwargs.pop("id", None)
        kwargs.pop("name", None)
        kwargs.pop("label", None)
        self.kwargs.update(kwargs)

        for attr, value in self.kwargs.items():
            attr = to_kebab_case(attr)
            kwargs[attr] = value

        return self.markup(**kwargs)
    
    def generate_string_attributes(self, **kwargs) -> str:
        """Generates a string with html attributes corresponding to the kwargs object"""
        return html_params(**kwargs)
    
    @abstractmethod
    def markup(self, **kwargs) -> Markup:
        """Must return the markuo of the element"""
        pass
    
    def label(self, **kwargs) -> Markup:
        """
        Generates the label markup for the form field.
        """
        if self._label is None:
            return Markup("")

        return Markup(f'<label for="{self.id}" class="{kwargs.get('class', '')}" style="{kwargs.get('style', '')}">{self._label}</label>')

class TagsInput(FormField):

    type: FormFieldTypes = FormFieldTypes.TAGS

    def __init__(self, **kwargs):
        """
        Initlizer for TagsInput element. Accepts standard html attributes
        and the special boolean attribute: "as_string". 
        If as_string=True, the return value of the html element will be a comma separated string,
        otherwise it will be a list of strings.
        """
        super().__init__(**kwargs)

    def markup(self, **kwargs) -> Markup:
        attrs = self.generate_string_attributes(**kwargs)
        return Markup(f'<tags-input id="{self.id}" name="{self.name}" {attrs}></tags-input>')

class PasswordInput(FormField):
    """
    Password input field
    """

    type: FormFieldTypes = FormFieldTypes.PASSWORD

    def markup(self, **kwargs) -> Markup:
        """
        Password field input
        """
        attrs = self.generate_string_attributes(**kwargs)
        return Markup(f'<input type="password" id="{self.id}" name="{self.name}" {attrs} />')

class EmailInput(FormField):
    """
    Email input field
    """

    type: FormFieldTypes = FormFieldTypes.EMAIL

    def markup(self, **kwargs) -> Markup:
        """
        Email field input
        """
        attrs = self.generate_string_attributes(**kwargs)
        return Markup(f'<input type="email" id="{self.id}" name="{self.name}" {attrs} />')

class TextInput(FormField):
    """
    Text input field
    """

    type: FormFieldTypes = FormFieldTypes.TEXT

    def markup(self, **kwargs) -> Markup:
        """
        Text field input
        """
        attrs = self.generate_string_attributes(**kwargs)
        return Markup(f'<input type="text" id="{self.id}" name="{self.name}" {attrs} />')

class TextAreaInput(FormField):
    """
    Text input field
    """

    type: FormFieldTypes = FormFieldTypes.TEXTAREA

    def markup(self, **kwargs) -> Markup:
        """
        Textarea field input
        """
        attrs = self.generate_string_attributes(**kwargs)
        return Markup(f'<textarea type="text" id="{self.id}" name="{self.name}" {attrs}></textarea>')

class SelectInput(FormField):
    """
    Select input field
    """

    type: FormFieldTypes = FormFieldTypes.SELECT

    def __init__(self, **kwargs):
        options = kwargs.pop("options", None)
        super().__init__(**kwargs)
        self.options = options
        if self.options is None or not isinstance(self.options, list) or not isinstance(self.options[0], tuple):
            raise Exception("Please provide a list of tuples (value, name) as options")
    
    def generate_options(self) -> str:
        """Generates options string"""
        options: str = ""
        for val, name in self.options:
            options+=f'<option value="{val}">{name}</options>'
        return options

    def markup(self, **kwargs) -> Markup:
        """
        Select field input
        """
        attrs = self.generate_string_attributes(**kwargs)
        options: str = self.generate_options()
        return Markup(f'<select type="text" id="{self.id}" name="{self.name}" {attrs}>{options}</select>')

class CheckboxInput(FormField):
    """
    Checkbox input field
    """

    type: FormFieldTypes = FormFieldTypes.CHECKBOX

    def markup(self, **kwargs) -> Markup:
        """
        Checkbox field input
        """
        attrs = self.generate_string_attributes(**kwargs)
        return Markup(f'<input type="checkbox" id="{self.id}" name="{self.name}" {attrs} />')

class RadioInput(FormField):
    """
    Radio input field
    """

    type: FormFieldTypes = FormFieldTypes.RADIO

    def markup(self, **kwargs) -> Markup:
        """
        Radio field input
        """
        attrs = self.generate_string_attributes(**kwargs)
        return Markup(f'<input type="radio" id="{self.id}" name="{self.name}" {attrs} />')



"""
Authentication module
"""

from .authentication import (login_required, role_required,
                                Authentication, AuthUtils, AuthConfig)

__all__ = ['login_required', 'role_required',
           'Authentication', 'AuthUtils', 'AuthConfig']



"""Researcher model"""
from sqlalchemy import Text
from sqlalchemy.orm import Mapped, mapped_column
from pyjolt.admin import register_model

from .base_model import DatabaseModel

@register_model
class Researcher(DatabaseModel):
    """Researcher database model"""
    __tablename__ = "researchers"

    fullname: Mapped[str] = mapped_column(Text, nullable=False)
    code: Mapped[str] = mapped_column(Text, nullable=False)
    email: Mapped[str] = mapped_column(Text, unique=True, nullable=False)
    education: Mapped[str] = mapped_column(Text, nullable=True)
    employment: Mapped[str] = mapped_column(Text, nullable=True)
    title: Mapped[str] = mapped_column(Text, nullable=True)



"""
State machine extension
"""
from __future__ import annotations
from abc import abstractmethod
from enum import Enum, IntEnum, StrEnum
from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Callable,
    Dict,
    NotRequired,
    Optional,
    Tuple,
    Type,
    TypedDict,
    cast,
)

from pydantic import BaseModel, ConfigDict, Field

from ..controller import Controller
from .state_machine_controller import StateMachineController, TransitionRequestData
from ..auth import login_required
from ..controller import path
from ..base_extension import BaseExtension
from ..request import Request
from ..response import Response

if TYPE_CHECKING:
    from pyjolt.pyjolt import PyJolt

# Any bound method (sync or async) -> async method returning Response
AsyncMethod = Callable[..., Awaitable[Response]]

class _StateMachineConfigs(BaseModel):
    """Configuration options for StateMachine extension"""
    model_config = ConfigDict(extra="allow")

    API_URL: Optional[str] = Field("/api/v1/state-machine",
                            description="API URL for state machine operations")
    INCLUDE_OPEN_API: Optional[bool] = Field(True,
                        description="Whether to include state machine endpoints in OpenAPI schema")
    USE_AUTH: Optional[bool] = Field(True, description="If the state machine requires user authentication")
    TRANSITION_DATA: Optional[Type[BaseModel]] = Field(TransitionRequestData,
                        description="Transition request pydantic model for validation")

class StateMachineConfig(TypedDict):
    """StateMachine extension configurations"""
    API_URL: NotRequired[str]
    INCLUDE_OPEN_API: NotRequired[bool]
    USE_AUTH: NotRequired[bool]
    TRANSITION_DATA: NotRequired[Type[BaseModel]]

def step_method(*steps: Enum|StrEnum):
    """Adds method as a step method for the given step in the state machine"""

    def decorator(func: AsyncMethod) -> AsyncMethod:
        # pylint: disable=protected-access
        func.__state_machine_step__ = list(steps)  # type: ignore[attr-defined]
        return func
    return decorator

class StateMachine(BaseExtension):
    """
    State machine extension class
    """

    def __init__(self, steps: Type[Enum|StrEnum|IntEnum], states: Type[Enum|StrEnum|IntEnum],
                 states_steps_map: dict[Any, dict[Any, Any]], configs_name: str = "STATE_MACHINE", ):
        self._app: "PyJolt" = cast("PyJolt", None)
        self._steps = steps
        self._states = states
        self._states_steps_map: dict[Any, dict[Any, Any]] = states_steps_map
        self._configs_name: str = configs_name
        self._configs: dict[str, Any] = cast(dict[str, Any], None)
        self._step_methods_map: Dict[Enum|StrEnum|IntEnum, Callable] = {}

    def init_app(self, app: "PyJolt") -> None:
        """
        Initialize the extension with the PyJolt app
        """
        self._app = app
        self._configs = cast(dict[str, Any], app.configs.get(self._configs_name, {}))
        if self._configs is None:
            raise ValueError(f"Configurations for {self._configs_name} not found in app configurations.")
        self._configs = self.validate_configs(self._configs, _StateMachineConfigs)
        self._get_step_methods()
        ctrl = path(self.configs["API_URL"])(StateMachineController)
        if self.configs.get("USE_AUTH", True):
            ctrl = cast(Type[Controller], login_required(ctrl))
        setattr(ctrl, "state_machine", self)
        self.app.register_controller(ctrl)

    def _get_step_methods(self) -> None:
        """Returns a dictionery with step methods mapping"""
        for name in dir(self):
            method = getattr(self, name)
            if not callable(method):
                continue
            steps = getattr(method, "__state_machine_step__", None)
            if steps:
                for step in steps:
                    if self._step_methods_map.get(step, None) is not None:
                        raise ValueError(f"Multiple methods found for step {step} in state machine.")
                    if step is not None:
                        self._step_methods_map[step] = method

    def get_next_state(self, current_state: Enum|StrEnum|IntEnum,
                step: Enum|StrEnum|IntEnum) -> Optional[Enum|StrEnum|IntEnum]:
        return self.states_steps_map.get(current_state, {}).get(step, None)

    @property
    def states_steps_map(self) -> dict[Any, dict[Any, Any]]:
        """
        Get the state step map
        """
        return self._states_steps_map

    @property
    def steps(self) -> Type[Enum|StrEnum|IntEnum]:
        return self._steps

    @property
    def states(self) -> Type[Enum|StrEnum|IntEnum]:
        return self._states

    @property
    def step_methods_map(self) -> Dict[Enum|StrEnum|IntEnum, Callable]:
        return self._step_methods_map

    @property
    def transition_request_data(self) -> Type[BaseModel]:
        return self._configs["TRANSITION_DATA"]

    @abstractmethod
    async def has_permission(self, req: Request) -> bool:
        """Checks if user has permission for action"""

    @abstractmethod
    async def context_loader(self, req: Request,
                    transition_request: Any) -> Tuple[Enum|IntEnum|StrEnum,Any]:
        """
        Loads the needed context for transitions. Should return a tuple with
        the current state (Enum|IntEnum|StrEnum) and any needed context
        """



"""State machine extension implementation"""
from enum import Enum, IntEnum, StrEnum
from typing import Any, Tuple

from pyjolt.http_statuses import HttpStatus
from pyjolt.request import Request
from pyjolt.response import Response
from pyjolt.state_machine import StateMachine as SM, step_method

class States(StrEnum):

    PENDING = "pend"
    UNDER_REVIEW = "urev"
    REJECTED = "rej"
    EVALUATED = "eval"

class Steps(StrEnum):

    ACCEPT_REVIEW = "accr"
    REJECT_REVIEW = "rejr"
    ACCEPT = "acc"
    REJECT = "rej"

STATE_STEP_MAP: dict[States, dict[Steps, States]] = {
    States.PENDING: {
        Steps.ACCEPT_REVIEW: States.UNDER_REVIEW,
        Steps.REJECT_REVIEW: States.REJECTED
    },
    States.UNDER_REVIEW: {
        Steps.REJECT: States.REJECTED,
        Steps.ACCEPT: States.EVALUATED
    }
}

class StateMachine(SM):
    """SM implementation"""

    @step_method(Steps.ACCEPT_REVIEW)
    async def on_accept_review(self, req: Request,
                               transition_request: Any,
                               context: Any,
                               next_state: Enum|IntEnum|StrEnum) -> Response:
        print("Performing step:", req, transition_request, context)
        return req.res.json({
            "message": f"Transition successful with step {transition_request.get('step').name} from state {transition_request.get("current_state").name} to state {next_state.name}",
            "status": "success"
        }).status(HttpStatus.OK)

    async def has_permission(self, req: Request) -> bool:
        return True

    async def context_loader(self, req: Request,
            transition_request: Any) -> Tuple[Enum | IntEnum | StrEnum, Any]:
        return States.PENDING, {}

state_machine: StateMachine = StateMachine(Steps, States, STATE_STEP_MAP)



"""Publication API"""
from time import time_ns
from typing import Any

from app.api.models.publication import Publication
from app.api.schemas.base_schema import ErrorResponseSchema
from app.api.schemas.publication_schemas import (
    PublicationOutSchema,
    PublicationResponseSchema,
    PublicationsQueryResponseSchema,
    PublicationsQuerySchema,
)
from app.extensions import db

from pyjolt import HttpStatus, MediaType, Request, Response, abort
from pyjolt.controller import Controller, Descriptor, get, open_api_docs, path, produces
from pyjolt.database.sql import AsyncSession


@path("/api/v1/publications")
class PublicationsApi(Controller):
    """Publications API controller"""

    @get("/")
    @produces(MediaType.APPLICATION_JSON)
    @db.managed_session
    async def get_publications(self, req: Request,
                session: AsyncSession) -> Response[PublicationsQueryResponseSchema]:
        """Gets all publications"""
        start_time: int = time_ns()
        query_data: PublicationsQuerySchema = PublicationsQuerySchema.model_validate(
                                                                        req.query_params)
        results: dict[str, Any] = await Publication.query_publications(session,
                                                                       query_data)
        end_time: int = time_ns()
        dur: float = (end_time - start_time)/1000000000.0
        self.app.logger.info(f"PERFORMANCE: {dur} s")
        return req.res.json({
            "message": "Query successful",
            "status": "success",
            "data": results
        }).status(HttpStatus.OK)

    @get("/<int:pub_id>")
    @open_api_docs(Descriptor(status=HttpStatus.NOT_FOUND,
                              description="Publication not found.",
                              body=ErrorResponseSchema))
    @produces(MediaType.APPLICATION_JSON)
    @db.managed_session
    async def get_publication_by_id(self, req: Request,
                                    pub_id: int,
                                    session: AsyncSession) -> Response[PublicationResponseSchema]:
        """Gets publication by id"""
        publication: Publication = await Publication.query(session).filter_by(id=pub_id).first()
        if publication is None:
            abort(f"Publication with id {pub_id} not found.", status_code=HttpStatus.NOT_FOUND)
        return req.res.json({
            "message": "Query successful",
            "status": "success",
            "data": PublicationOutSchema.from_model(publication)
        }).status(HttpStatus.OK)

    @get("/<path:doi>")
    @open_api_docs(Descriptor(status=HttpStatus.NOT_FOUND,
                              description="Publication not found.",
                              body=ErrorResponseSchema))
    @produces(MediaType.APPLICATION_JSON)
    @db.managed_session
    async def get_publication_by_doi(self, req: Request,
                                     doi: str,
                                     session: AsyncSession) -> Response[PublicationResponseSchema]:
        """Gets publication by DOI"""
        publication: Publication = await Publication.query(session).filter_by(doi=doi).first()
        if publication is None:
            abort(f"Publication with DOI {doi} not found.", status_code=HttpStatus.NOT_FOUND)
        return req.res.json({
            "message": "Query successful",
            "status": "success",
            "data": PublicationOutSchema.from_model(publication)
        }).status(HttpStatus.OK)



"""
Custom exception classes for PyJolt
"""
from typing import Type

class CustomException(Exception):
    """Base custom exception class"""

class MethodNotControllerMethod(CustomException):
    """
    Error if the decorated method is not part of a controller
    """
    def __init__(self, message):
        self.message = message

class UnexpectedDecorator(CustomException):
    """
    Error for unexpected decorators
    """
    def __init__(self, message):
        self.message = message

class MissingDecoratorError(CustomException):
    """
    Error for missing decorators
    """
    def __init__(self, message):
        self.message = message

class DuplicateRoutePath(CustomException):
    """
    Error for duplicate route path
    """

    def __init__(self, message):
        self.message = message

class DuplicateExceptionHandler(CustomException):
    """
    Error for duplicate registered exception handler
    """
    def __init__(self, message):
        self.message = message

class Jinja2NotInitilized(CustomException):
    """
    Error if jinja2 is not initilized
    """
    def __init__(self):
        self.message = "Jinja2 render engine is not initilized."

class MissingExtension(CustomException):
    """
    Error for missinf extension
    """
    def __init__(self, ext_name: str):
        self.message = f"Extension with name {ext_name} not found on application."

class MissingDependencyInjectionMethod(CustomException):
    """
    Error for missing dependency injection method.
    app._dependency_injection_map.get() returns None
    """
    def __init__(self, injectable: Type):
        self.message = f"Missing dependency injection method. Please provide a dependency injection method for the dependency: {injectable.__name__}"

class MissingResponseObject(CustomException):
    """
    Error if route handler method does not return a response object (pyjolt.Response)
    """
    def __init__(self):
        self.message = "Route handler must return a result of type pyjolt.Response"

class MissingRouterInstance(CustomException):
    """
    Error if ClassBlueprint does not have router instances defined.
    """
    def __init__(self):
        self.message = """Missing ClassBlueprint router instance. Did you forget 
                        to add the @controller decorator on the ClassBlueprint?"""

class InvalidRouteHandler(CustomException):
    """
    Error for invalid route handlers in ClassBlueprint
    Missing "is_route_handler" attribute
    """
    def __init__(self):
        self.message = """This is not a valid route handler method for ClassBlueprint"""

class InvalidWebsocketHandler(CustomException):
    """
    Error for invalid websocket handlers in ClassBlueprint
    Missing "is_websocket_handler" attribute
    """
    def __init__(self):
        self.message = """This si not a valid websocket handler method for ClassBlueprint"""



"""
Example data model
"""
from typing import Type
from sqlalchemy import String
from sqlalchemy.orm import mapped_column, Mapped

from pyjolt.database.sql import create_declarative_base, DeclarativeBaseModel

Base: Type[DeclarativeBaseModel] = create_declarative_base()

class Example(Base):#type: ignore[valid-type,misc]
    """
    Example model
    """
    #table name in database; usually plural
    __tablename__: str = "examples"

    id: Mapped[int] = mapped_column(primary_key=True)
    name: Mapped[str] = mapped_column(String(30))



"""
File controller for pyjolt admin dashboard
"""
import os
import io
import zipfile
import uuid
from typing import Any, Optional
import mimetypes
from pathlib import Path
import shutil
from ..http_statuses import HttpStatus
from ..controller import get, post, delete
from .common_controller import CommonAdminController
from ..auth import login_required
from ..request import Request
from ..response import Response
from ..utilities import get_file, fs_safe_join

class AdminFileController(CommonAdminController):

    @get("/files")
    @login_required
    async def file_explorer(self, req: Request) -> Response:
        """
        File explorer page
        """
        await self.can_enter(req)
        if not await self.dashboard.has_files_permission(req):
            return await self.missing_files_permission(req)
        return await req.res.html("/__admin_templates/file_explorer.html", {
            "folder": self.app.get_conf("STATIC_DIR"),
            "files_and_folders": await self.get_files_and_folder(None),
            **self.get_common_variables()
        })
    
    @get("/files/fetch")
    @login_required
    async def get_all(self, req: Request) -> Response:
        """
        Returns all files
        """
        folder_path: Optional[str] = req.query_params.get("path", None)
        return req.res.json({
            "message": "Files fetched successfully",
            "status": "success",
            "data": await self.get_files_and_folder(folder_path)
        }).status(HttpStatus.OK)
    
    @post("/files/fetch/zip")
    @login_required
    async def download_zip(self, req: Request) -> Response:
        """Downloads multiple files as zip"""
        await self.can_enter(req)
        if not await self.dashboard.has_files_permission(req):
            return await self.missing_files_permission(req)
        req_items: Optional[dict[str, Any]] = await req.get_data()
        if req_items is None:
            return req.res.json({
                "message": "Please provide a valid list of files for download",
                "status": "error"
            }).status(HttpStatus.BAD_REQUEST)
        allowed_base = Path(self.app.root_path).resolve()

        zip_buffer = io.BytesIO()
        with zipfile.ZipFile(zip_buffer, "w", zipfile.ZIP_DEFLATED) as zf:
            for name, rel_path in req_items.items():
                try:
                    target_path = fs_safe_join(allowed_base, rel_path.lstrip("/\\"), name)
                    if target_path is None:
                        raise FileNotFoundError()
                    target = Path(target_path).resolve()
                    if target.is_dir():
                        top = Path(name).name
                        zf.writestr(f"{top}/", b"")
                        for p in target.rglob("*"):
                            if p.is_dir():
                                continue
                            rel_inside = p.relative_to(target)
                            arcname = (Path(top) / rel_inside).as_posix()
                            zf.write(p, arcname=arcname)
                    elif target.is_file():
                        arcname = Path(name).name
                        zf.write(target, arcname=arcname)
                    else:
                        raise FileNotFoundError()
                except FileNotFoundError:
                    self.app.logger.info(f"Requested item not found: {name=} {rel_path=}")
                    return req.res.json({
                        "message": f"Requested item '{name=}' does not exist.",
                        "status": "danger"
                    }).status(HttpStatus.NOT_FOUND)
        zip_buffer.seek(0)
        return req.res.send_file(zip_buffer.read(), {
            "Content-Type": "application/zip",
            "Content-Disposition": f"attachment; filename={uuid.uuid4().hex}.zip"
        })
    
    @get("/files/fetch/one")
    @login_required
    async def get_file(self, req: Request) -> Response:
        """
        Returns selected file
        """
        file_path: Optional[str] = req.query_params.get("path", None)
        if file_path is None:
            return req.res.json({
                "message": "Please provide a valid file path",
                "status": "warning"
            }).status(HttpStatus.BAD_REQUEST)
        full_path = os.path.join(self.app.root_path, file_path.lstrip("/\\"))
        if not os.path.isfile(full_path):
            return req.res.json({
                "message": "File does not exist",
                "status": "danger"
            }).status(HttpStatus.BAD_REQUEST)
        guessed, _ = mimetypes.guess_type(full_path)
        content_type = guessed or "application/octet-stream"
        status, headers, body = await get_file(full_path, content_type=content_type)
        headers["Accept-Ranges"] = "bytes"
        return req.res.send_file(body, headers).status(status)
    
    @get("/files/rename")
    @login_required
    async def rename(self, req: Request) -> Response:
        """
        Renames file or folder
        """
        await self.can_enter(req)
        if not await self.dashboard.has_files_permission(req):
            return await self.missing_files_permission(req)
        data: dict[str, str] = {}
        for param in ["path", "oldName", "newName"]:
            param_data: Optional[str] = req.query_params.get(param, None)
            if param_data is None:
                return req.res.json({
                    "message": f"Missing parameter '{param}'",
                    "status": "warning"
                }).status(HttpStatus.BAD_REQUEST)
            data[param] = param_data
        
        data["path"] = os.path.join(self.app.root_path, data["path"].lstrip("/\\"))
        old_path = os.path.join(data["path"], data["oldName"])
        new_path = os.path.join(data["path"], data["newName"])
        print(data["path"], old_path, new_path)
        try:
            os.rename(old_path, new_path)
        except Exception:
            return req.res.json({
                "message": f"Failed to change file name {data['oldName']}",
                "status": "danger"
            }).status(HttpStatus.BAD_REQUEST)

        return req.res.json({
            "message": "File or folder renamed successfully",
            "status": "success",
            "data": {
                "old_path": old_path,
                "new_path": new_path,
                "new_name": data["newName"],
                "old_name": data["oldName"]
            }
        }).status(HttpStatus.OK)
    
    @post("/files/upload")
    @login_required
    async def upload(self, req: Request) -> Response:
        """
        Uploads single file
        """
        await self.can_enter(req)
        if not await self.dashboard.has_files_permission(req):
            return await self.missing_files_permission(req)
        data: dict[str, Any] = await req.form_and_files()
        full_path = os.path.join(self.app.root_path, data["path"].lstrip("/\\"))
        file_path = Path(full_path)
        file_path.parent.mkdir(parents=True, exist_ok=True)
        file_path.write_bytes(data["file"].read())
        return req.res.json({
            "message": "File uploaded successfully",
            "status": "success",
            "data": {
                "path": data["path"],
                "name": data["file"].filename,
                "is_folder": False
            }
        }).status(HttpStatus.CREATED)
    
    @delete("/files")
    @login_required
    async def delete(self, req: Request) -> Response:
        """
        Delete a file or folder
        """
        await self.can_enter(req)
        if not await self.dashboard.has_files_permission(req):
            return await self.missing_files_permission(req)
        path: Optional[str] = req.query_params.get("path", None)
        full_path = os.path.join(self.app.root_path, path.lstrip("/\\") if path is not None else "")
        if path is None or (not os.path.isfile(full_path) and not os.path.isdir(full_path)):
            return req.res.json({
                "message": "Please provide a valid file or folder path",
                "status": "danger"
            }).status(HttpStatus.BAD_REQUEST)

        
        if os.path.isdir(full_path):
            shutil.rmtree(full_path)
        else:
            os.remove(full_path)

        return req.res.no_content()

    async def get_files_and_folder(self, path: Optional[str]) -> list[dict[str, str|bool|None]]:
        """Returns all files and folders at the provided path"""
        if path is None:
            path = self.app.static_files_path
        else:
            if path.startswith("/static"):
                path = path.replace("/static", "")
            elif path.startswith("static"):
                path = path.replace("static", "")
            path = os.path.join(self.app.static_files_path, *path.split("/"))
        items = os.listdir(path)
        files_and_folders = []
        for name in items:
            full = os.path.join(path, name)
            guessed, _ = mimetypes.guess_type(full)
            files_and_folders.append({
                "path": path.replace(self.app.root_path, ""),
                "name": name,
                "is_folder": True if os.path.isdir(full) else False,
                "mimetype": guessed
            })
        
        return files_and_folders
    
    async def missing_files_permission(self, req: Request) -> Response:
        """Returns missing files permission"""
        return await req.res.html(
            "/__admin_templates/missing_permission.html"
        )



"""
Ai interface module
"""
from .ai_interface import AiInterface, ChatContextNotFound, tool, FailedToRunAiToolMethod, AiConfig

__all__ = ['AiInterface', 'ChatContextNotFound', 'tool', 'FailedToRunAiToolMethod', 'AiConfig']



"""Project model"""
from typing import Any
from sqlalchemy import Text
from sqlalchemy.orm import Mapped, mapped_column
from pyjolt.database.sql import AsyncSession
from pyjolt.admin import register_model

from app.api.schemas.project_schemas import ProjectQuerySchema, ProjectOutSchema
from .base_model import DatabaseModel

@register_model
class Project(DatabaseModel):
    """Project db model"""

    __tablename__ = "projects"

    title_slv: Mapped[str] = mapped_column(Text, nullable=False)
    title_eng: Mapped[str] = mapped_column(Text, nullable=True)
    leader: Mapped[str] = mapped_column(Text, nullable=False)
    duration: Mapped[str] = mapped_column(Text, nullable=False)
    sicris_url: Mapped[str] = mapped_column(Text, nullable=False)
    code: Mapped[str] = mapped_column(Text, nullable=False)
    abstract_slv: Mapped[str] = mapped_column(Text, nullable=True)
    abstract_eng: Mapped[str] = mapped_column(Text, nullable=True)
    keywords_slv: Mapped[str] = mapped_column(Text, nullable=False)
    keywords_eng: Mapped[str] = mapped_column(Text, nullable=True)
    researchers: Mapped[str] = mapped_column(Text, nullable=False)

    @classmethod
    async def query_projects(cls, session: AsyncSession,
                             query_data: "ProjectQuerySchema") -> dict[str, Any]:
        """Queries projects"""
        conds = []
        if query_data.title_slv is not None:
            conds.append(query_data.title_slv in cls.title_slv)
        if query_data.title_eng is not None:
            conds.append(query_data.title_eng in cls.title_eng)
        if query_data.end_date is not None:
            conds.append(query_data.end_date.strftime("%YYYY-%MM-%dd") in cls.duration)
        if query_data.start_date is not None:
            conds.append(query_data.start_date.strftime("%YYYY-%MM-%dd") in cls.duration)
        if query_data.keywords_slv is not None:
            conds.append(query_data.keywords_slv in cls.keywords_slv)
        if query_data.keywords_eng is not None:
            conds.append(query_data.keywords_eng in cls.keywords_eng)
        if query_data.leader is not None:
            conds.append(query_data.leader in cls.leader)
        if query_data.researchers is not None:
            rsrs: list[str] = query_data.researchers.split(",")
            for rsr in rsrs:
                conds.append(rsr.strip() in cls.researchers)
        projects: dict[str, Any] = await cls.query(session).filter(
            *conds
        ).paginate(page=query_data.page, per_page=query_data.per_page)
        projects["items"] = [ProjectOutSchema.from_model(project)
                          for project in projects["items"]]
        return projects
